/* 
   Program generate triangulations of 3-manifolds by adding tetrahedra in 
   lexigraphic order.
*/

#define VERSION "0.26"

/* use look-ahead to force additional tetrahedron to reduce choice.s */
/* store two triangulations. _a for all added tetrahedra.  _p for all picked 
   tetrahedra. */

/* possible (mutually exclusive) compile macros
   MAX5  maximum edge degree is 5.
   NO66  edge degree is 5 or 6 and no face has two edges with edge degree 6.
   DODEC  NO66 requiring one vertex to have degree 12.
   NEIGHBORLY  all vertices are adjacent.
*/

#if defined DODEC

#define NO66 1

#endif

#if defined MAX5

#define MIN_VERTEX_DEGREE 4
#define MAX_VERTEX_DEGREE 12
#define MIN_EDGE_DEGREE 3
#define MAX_EDGE_DEGREE 5
#define TCP 1

#elif defined NO66

#define MIN_VERTEX_DEGREE 12
#define MAX_VERTEX_DEGREE 16
#define MIN_EDGE_DEGREE 5
#define MAX_EDGE_DEGREE 6
#define TCP 1

#endif

#define USAGE \
" lextet [-v] [-h] [-i] [-o o] [-r res -m mod] nv"

#define HELPTEXT \
" lextet : generate triangulations of 3-manifolds with nv vertices.\n\
\n\
 help switches:\n\
   -h  show this text\n\
   -v  verbose output\n\
\n\
 parameter switches:\n\
   -i          if present only irreducible triangulations are generated\n\
   -o o        0: only non-orientable, 1: only orientable, otherwise: both\n\
   -r res      res for splitting\n\
   -m mod      mod for splitting\n\
\n\
   nv          number of vertices in generated triangulations.\n"

#include <stdio.h>
#include <string.h>

#if __STDC__
#include <stdlib.h>
#endif

#define CPUTIME 1          /* Whether to measure the cpu time or not */

#if CPUTIME
#include <sys/times.h>
#include <time.h>
#if !defined(CLK_TCK) && !defined(_SC_CLK_TCK)
#include <unistd.h>
#endif
#if !defined(CLK_TCK) && defined(_SC_CLK_TCK)
#define CLK_TCK sysconf(_SC_CLK_TCK)
#endif
#if !defined(CLK_TCK) && defined(CLOCKS_PER_SEC)
#define CLK_TCK CLOCKS_PER_SEC
#endif
#if !defined(CLK_TCK)
#define CLK_TCK 60     /* If the CPU time stated by the program appears
                      to be out by a constant ratio, the most likely
                      explanation is that the code got to this point but
                      60 is the wrong guess.  Another common value is 100. */
#endif
#endif

#define FALSE 0
#define TRUE 1

#define MAXN 200
#define MAXE 8400
#define MAXF MAXN*(MAXN-1)*(MAXN-2)/(3*2)
#define MAXT 8400

/* BIG macros taken from plantri. */
/* The program is so fast that the count of output graphs can quickly
   overflow a 32-bit integer.  Therefore, we use two long values
   for each count, with a ratio of 10^9 between them.  The macro
   ADDBIG adds a small number to one of these big numbers.  
   BIGTODOUBLE converts a big number to a double (approximately).
   SUMBIGS adds a second big number into a first big number.
   SUBBIGS subtracts a second big number from a first big number.
   PRINTBIG prints a big number in decimal.
   ZEROBIG sets the value of a big number to 0.
   ISZEROBIG tests if the value is 0.
   SETBIG sets a big number to a value at most 10^9-1.
   ISEQBIG tests if two big numbers are equal.
*/

typedef struct
{
    long hi,lo;
} bigint;

#define ZEROBIG(big) big.hi = big.lo = 0L
#define ISZEROBIG(big) (big.lo == 0 && big.hi == 0)
#define SETBIG(big,value) {big.hi = 0L; big.lo = (value);}
#define ADDBIG(big,extra) if ((big.lo += (extra)) >= 1000000000L) \
    { ++big.hi; big.lo -= 1000000000L;}
#define PRINTBIG(file,big) if (big.hi == 0) \
 fprintf(file,"%ld",big.lo); else fprintf(file,"%ld%09ld",big.hi,big.lo)
#define BIGTODOUBLE(big) (big.hi * 1000000000.0 + big.lo)
#define SUMBIGS(big1,big2) {if ((big1.lo += big2.lo) >= 1000000000L) \
    {big1.lo -= 1000000000L; big1.hi += big2.hi + 1L;} \
    else big1.hi += big2.hi;}
#define SUBBIGS(big1,big2) {if ((big1.lo -= big2.lo) < 0L) \
    {big1.lo += 1000000000L; big1.hi -= big2.hi + 1L;} \
    else big1.hi -= big2.hi;}
/* Note: SUBBIGS must not allow the value to go negative.
   SUMBIGS and SUBBIGS both permit big1 and big2 to be the same bigint. */
#define ISEQBIG(big1,big2) (big1.lo == big2.lo && big1.hi == big2.hi)


typedef struct
{
  int other_a[2];
  int type_a[2];  /* 0 = available, 1 = forced, 2 = picked, 
		   3 = forced then picked */
  int other_p[2];
#ifdef COVER
  int other_c[2];
#endif
} face;

#define AVAILABLE 0
#define FORCED 1
#define PICKED 2
#define FORCED_PICKED 3

typedef struct
{
  int mark;            /* for temporary use;
                          Only access mark via the MARK macros. */
} vert;

typedef struct
{
  int mark;            /* for temporary use;
                          Only access mark via the MARK macros. */
} oface;

static int markvalue = 30000;
#define RESETMARKS {int mki; if ((markvalue += 1) > 30000) \
       { markvalue = 1; for (mki=0;mki<2*maxnf;++mki) ofaces[mki].mark=0; \
         for (mki=0;mki<maxnv;++mki) verts[mki].mark=0;}}
#define MARK(e) (e)->mark = markvalue
#define ISMARKED(e) ((e)->mark >= markvalue)

static face faces[MAXF];
static face *base[MAXN][MAXN][MAXN];

static vert verts[MAXN];
static vert *vertspt[MAXN];
static oface ofaces[2*MAXF];
static oface *ofacespt[MAXN][MAXN][MAXN];

static int verbose;
static int only_irreducible; /* flag if only irreducible triangulations are to
				be generated */
static int nv_a;     /* number of vertices in final triangulation */
static int ne_a;     /* number of edges in triangulation */
static int nf_a;     /* number of faces in triangulation */
static int nt_a;     /* number of tetrahedra in triangulation */
static int nv_p;     /* number of vertices in final triangulation */
static int ne_p;     /* number of edges in triangulation */
static int nf_p;     /* number of faces in triangulation */
static int nt_p;     /* number of tetrahedra in triangulation */
static int maxnv;  /* maximum number of vertices in a triangulation */
static int maxne;  /* maximum number of edges in a triangulation */
static int maxnf;  /* maximum number of faces in a triangulation */
static int maxnt;  /* maximum number of tetrahedra in a triangulation */

/* degreeMN is number of N-complexes containing M-complex, M < N */

static int degree01_a[MAXN]; /* edge degree of vertices, normal degree */
static int degree02_a[MAXN]; /* face degree at vertex */
static int degree03_a[MAXN]; /* tetrahedron degree of vertices, number of 
			     tetrahedra with that vertex */
static int degree12_a[MAXN][MAXN]; /* face degree around an edge 
				    = # of vertices in lk of edge */
static int degree13_a[MAXN][MAXN]; /* tetrahedron degree around an edge 
				    = # edge in lk of edge */
static int degree01_p[MAXN]; /* edge degree of vertices, normal degree */
static int degree02_p[MAXN]; /* face degree at vertex */
static int degree03_p[MAXN]; /* tetrahedron degree of vertices, number of 
			     tetrahedra with that vertex */
static int degree12_p[MAXN][MAXN]; /* face degree around an edge 
				    = # of vertices in lk of edge */
static int degree13_p[MAXN][MAXN]; /* tetrahedron degree around an edge 
				    = # edge in lk of edge */
static int complete_a[MAXN]; /* link of vertex is a triangulation of a sphere */
static int complete_p[MAXN]; /* link of vertex is a triangulation of a sphere */
static int completeness[MAXN]; /* 0 if not complete, 1 if newly complete, or 
				  2 if formerly complete */

static int match[MAXN][120][MAXN]; /* lk(v0) is equivalent to lk(0),
				      i is an index of the automorphisms of 
				      this equivalence,
				      v is mapped to match[v0][i][v] by the 
				      i-th automorphism */
static int nmatch[MAXN];   /* number of automorphisms above */
static int list_a[MAXT][4];  /* list of the vertices in each ordered 
			      tetrahedron as added to _a */
static int list_a_forced[MAXT]; /* the index of the picked tetrahedron which
				   forced this tetrahedron to be added; 
				   -1 if not forced */
static int list_p[MAXT][4];  /* list of the vertices in each ordered 
			      tetrahedron as added to _p */

static bigint ngenerated[MAXN]; /* number of triangulations generated by ne */
static bigint ngen_all; /* number of triangulations generated */
static int nt_match;        /* number of tetrahedra matched */
static int v_to_label[MAXN];/* relabeling of vertices */
static int label_to_v[MAXN];/* vertex with label*/
static int next_label;      /* next unused label */
static int debug_count;

static bigint ncalls_min_lex;
static int max_inter_nv;
static bigint closed_link[MAXN]; /* closed_link[i] = number of times i-th 
				    vertex link is closed */
static int prev_closed_link;
static int res,mod;        /* res/mod from command line (default 0/1) */
static char res_text[10];  /* text string for res with padded zeros */
static int splitlevel,
           splitcount;     /* used for res/mod splitting */
static int orientability;  /* orientability of output triangulations, 
			      0 for only non-orientable, 1 for orientable, 
			      otherwise both */
static FILE *outfile_alpha[MAXE];
static FILE *outfile_lex[MAXE];

#define MAX(x,y) ((x)<(y) ? (y) : (x))
#define MIN(x,y) ((x)>(y) ? (y) : (x))

#ifdef SPLITTEST
static int splitcases;
#endif

#ifdef COVER
static nv_c;       /* number of vertices in triangulation to be covered */
static multi_c;    /* multiplicity of covering */
static cm[MAXT][MAXN];   
                   /* cover map: cm[v] is the vertex in the base triangulation
		      into which v maps */
static cmult[MAXT][MAXN]; /* cover multiplicity */
static degree0_c;  /* degree of edge (0,1) in triangulation to be covered */
static int degree13_c[MAXN][MAXN];
#endif

void found_one();
void next_tetrahedron();

void error_exit(int errornum)
{
  fprintf(stderr,"Error number %d, exiting\n",errornum);
  exit(1);
}

static char i2a[52] = {
  'a','b','c','d','e','f','g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y','z',
  'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q','R','S','T','U','V','W','X','Y','Z'};

static int a2i[128] = {
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50,51,-1,-1,-1,-1,-1,
  -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,-1,-1,-1,-1,-1};

void write_lex(FILE *output)
{
  /* write out 3-manifold in sortable format.  
     one line of text for each 3-manifold.
     one character for each vertex of each tetrahedron. */

  int itet,v;

  if (nv_p <= 52) 
    for (itet=0; itet<nt_p; itet++)
      fprintf(output,"%c%c%c%c",
	      i2a[list_p[itet][0]],i2a[list_p[itet][1]],i2a[list_p[itet][2]],
	      i2a[list_p[itet][3]]);
  else
    for (itet=0; itet<nt_p; itet++)
      fprintf(output,"[%d,%d,%d,%d],\n",
	      list_p[itet][0]+1,list_p[itet][1]+1,list_p[itet][2]+1,
	      list_p[itet][3]+1);
  fprintf(output,"\n");

}

#ifdef COVER
int read_lex()
{
  /* read a 3-manifold from input in lex format. */

#define LINELEN 2000
  char text[LINELEN];

  int v0,v1,v2,v3;
  int nv_read;
  int ichar,i,j,v[4];

  /* read the lex representation of the triangulation */

  if (fgets(text,LINELEN,stdin) == NULL) {
    return 0;
  }

  for (v0=0; v0<maxnv; v0++)
    for (v1=0; v1<maxnv; v1++)
      degree13_c[v0][v1] = 0;
  degree0_c = 0;
  nv_read = 0;
  ichar = 0;
  while (ichar+4 < strlen(text)) {
    for (i=0; i<4; i++)
      v[i] = a2i[text[ichar++]];
    if (v[3]+1 > nv_read)
      nv_read = v[3]+1;
    if (nv_read <= maxnv) {

      for (i=0; i<3; i++)
	for (j=i+1; j<4; j++) {
	  degree13_c[v[i]][v[j]]++;
	  degree13_c[v[j]][v[i]]++;
	}

      v0 = v[0]; v1 = v[1]; v2 = v[2]; v3 = v[3];
      if (base[v1][v2][v3]->other_c[0] == -1)
	base[v1][v2][v3]->other_c[0] = v0;
      else
	base[v1][v2][v3]->other_c[1] = v0;
      
      v0 = v[1]; v1 = v[0]; v2 = v[2]; v3 = v[3];
      if (base[v1][v2][v3]->other_c[0] == -1)
	base[v1][v2][v3]->other_c[0] = v0;
      else
	base[v1][v2][v3]->other_c[1] = v0;
      
      v0 = v[2]; v1 = v[0]; v2 = v[1]; v3 = v[3];
      if (base[v1][v2][v3]->other_c[0] == -1)
	base[v1][v2][v3]->other_c[0] = v0;
      else
	base[v1][v2][v3]->other_c[1] = v0;
      
      v0 = v[3]; v1 = v[0]; v2 = v[1]; v3 = v[2];
      if (base[v1][v2][v3]->other_c[0] == -1)
	base[v1][v2][v3]->other_c[0] = v0;
      else
	base[v1][v2][v3]->other_c[1] = v0;
    }
  }

  degree0_c = degree13_c[0][1];

  return nv_read;
}
 
#endif

void check_it()
{
  int ilist;
  int v0,v1,v2,v3;
  int lnv,lne,lnf,lnt;
  int gre01[MAXN];
  int gre02[MAXN];
  int gre03[MAXN];
  int gre12[MAXN][MAXN];
  int gre13[MAXN][MAXN];
  
  int is_v,is_e;
  
  for (v0=0; v0<maxnv; v0++) {
    gre01[v0] = 0;
    gre02[v0] = 0;
    gre03[v0] = 0;
    for (v1=0; v1<maxnv; v1++) {
      gre12[v0][v1] = 0;
      gre13[v0][v1] = 0;
    }
  }
  
  lnv = 0;
  lne = 0;
  lnf = 0;
  lnt = 0;
  
  for (ilist=0; ilist<nt_a; ilist++) {
    if (base[list_a[ilist][0]][list_a[ilist][1]][list_a[ilist][2]]->other_a[0]
	!= list_a[ilist][3] &&
	base[list_a[ilist][0]][list_a[ilist][1]][list_a[ilist][2]]->other_a[1]
	!= list_a[ilist][3])
      error_exit(101);
    if (base[list_a[ilist][1]][list_a[ilist][2]][list_a[ilist][3]]->other_a[0]
	!= list_a[ilist][0] &&
	base[list_a[ilist][1]][list_a[ilist][2]][list_a[ilist][3]]->other_a[1]
	!= list_a[ilist][0])
      error_exit(102);
    if (base[list_a[ilist][2]][list_a[ilist][3]][list_a[ilist][0]]->other_a[0]
	!= list_a[ilist][1] &&
	base[list_a[ilist][2]][list_a[ilist][3]][list_a[ilist][0]]->other_a[1]
	!= list_a[ilist][1])
      error_exit(103);
    if (base[list_a[ilist][3]][list_a[ilist][0]][list_a[ilist][1]]->other_a[0]
	!= list_a[ilist][2] &&
	base[list_a[ilist][3]][list_a[ilist][0]][list_a[ilist][1]]->other_a[1]
	!= list_a[ilist][2])
      error_exit(104);
  }
  
  for (v0=0; v0<nv_a; v0++) 
    if (complete_a[v0] != (2*degree02_a[v0] == 3*degree03_a[v0]))
      error_exit(106);

  for (v0=0; v0<nv_a; v0++) {
    is_v = FALSE;
    for (v1=0; v1<nv_a; v1++) {
      is_e = FALSE;
      if (v1 != v0)
	for (v2=0; v2<nv_a; v2++)
	  if (v2 != v0 && v2 != v1)
	    if (base[v0][v1][v2]->other_a[0] != -1) {
	      is_v = TRUE;
	      is_e = TRUE;
	      lnf++;
	      gre02[v0]++;
	      gre02[v1]++;
	      gre02[v2]++;
	      gre12[v0][v1]++;
	      gre12[v0][v2]++;
	      gre12[v1][v0]++;
	      gre12[v1][v2]++;
	      gre12[v2][v0]++;
	      gre12[v2][v1]++;
	      
	      lnt++;
	      v3 = base[v0][v1][v2]->other_a[0];
	      gre03[v0]++;
	      gre03[v1]++;
	      gre03[v2]++;
	      gre03[v3]++;
	      gre13[v0][v1]++;
	      gre13[v0][v2]++;
	      gre13[v0][v3]++;
	      gre13[v1][v0]++;
	      gre13[v1][v2]++;
	      gre13[v1][v3]++;
	      gre13[v2][v0]++;
	      gre13[v2][v1]++;
	      gre13[v2][v3]++;
	      gre13[v3][v0]++;
	      gre13[v3][v1]++;
	      gre13[v3][v2]++;
	      if (base[v0][v1][v2]->other_a[1] != -1) {
		lnt++;
		v3 = base[v0][v1][v2]->other_a[1];
		gre03[v0]++;
		gre03[v1]++;
		gre03[v2]++;
		gre03[v3]++;
		gre13[v0][v1]++;
		gre13[v0][v2]++;
		gre13[v0][v3]++;
		gre13[v1][v0]++;
		gre13[v1][v2]++;
		gre13[v1][v3]++;
		gre13[v2][v0]++;
		gre13[v2][v1]++;
		gre13[v2][v3]++;
		gre13[v3][v0]++;
		gre13[v3][v1]++;
		gre13[v3][v2]++;
	      }
	    }
      if (is_e) {
	lne++;
	gre01[v0]++;
	gre01[v1]++;
      }
    }
    if (is_v)
      lnv++;
  }
  
  if (lnv != nv_a) 
    error_exit(111);
  if (lne != 2*ne_a) 
    error_exit(112);
  if (lnf != 3*2*nf_a) 
    error_exit(113);
  if (lnt != 4*3*2*nt_a) 
    error_exit(114);
  
  for (v0=0; v0<maxnv; v0++) {
    if (gre01[v0] != 2*degree01_a[v0])
      error_exit(121);
    if (gre02[v0] != 3*2*degree02_a[v0])
      error_exit(122);
    if (gre03[v0] != 4*3*2*degree03_a[v0])
      error_exit(123);
    for (v1=0; v1<maxnv; v1++) {
      if (gre12[v0][v1] != 3*2*degree12_a[v0][v1])
	error_exit(124);
      if (gre13[v0][v1] != 4*3*2*degree13_a[v0][v1])
	error_exit(125);
    }
  }
  
  /* now for _p */

  for (v0=0; v0<maxnv; v0++) {
    gre01[v0] = 0;
    gre02[v0] = 0;
    gre03[v0] = 0;
    for (v1=0; v1<maxnv; v1++) {
      gre12[v0][v1] = 0;
      gre13[v0][v1] = 0;
    }
  }
  
  lnv = 0;
  lne = 0;
  lnf = 0;
  lnt = 0;
  
  for (ilist=0; ilist<nt_p; ilist++) {
    if (base[list_p[ilist][0]][list_p[ilist][1]][list_p[ilist][2]]->other_p[0]
	!= list_p[ilist][3] &&
	base[list_p[ilist][0]][list_p[ilist][1]][list_p[ilist][2]]->other_p[1]
	!= list_p[ilist][3])
      error_exit(201);
    if (base[list_p[ilist][1]][list_p[ilist][2]][list_p[ilist][3]]->other_p[0]
	!= list_p[ilist][0] &&
	base[list_p[ilist][1]][list_p[ilist][2]][list_p[ilist][3]]->other_p[1]
	!= list_p[ilist][0])
      error_exit(202);
    if (base[list_p[ilist][2]][list_p[ilist][3]][list_p[ilist][0]]->other_p[0]
	!= list_p[ilist][1] &&
	base[list_p[ilist][2]][list_p[ilist][3]][list_p[ilist][0]]->other_p[1]
	!= list_p[ilist][1])
      error_exit(203);
    if (base[list_p[ilist][3]][list_p[ilist][0]][list_p[ilist][1]]->other_p[0]
	!= list_p[ilist][2] &&
	base[list_p[ilist][3]][list_p[ilist][0]][list_p[ilist][1]]->other_p[1]
	!= list_p[ilist][2])
      error_exit(204);
  }
  
  for (v0=0; v0<nv_p; v0++) 
    if (complete_p[v0] != (2*degree02_p[v0] == 3*degree03_p[v0]))
      error_exit(206);

  for (v0=0; v0<nv_p; v0++) {
    is_v = FALSE;
    for (v1=0; v1<nv_p; v1++) {
      is_e = FALSE;
      if (v1 != v0)
	for (v2=0; v2<nv_p; v2++)
	  if (v2 != v0 && v2 != v1)
	    if (base[v0][v1][v2]->other_p[0] != -1) {
	      is_v = TRUE;
	      is_e = TRUE;
	      lnf++;
	      gre02[v0]++;
	      gre02[v1]++;
	      gre02[v2]++;
	      gre12[v0][v1]++;
	      gre12[v0][v2]++;
	      gre12[v1][v0]++;
	      gre12[v1][v2]++;
	      gre12[v2][v0]++;
	      gre12[v2][v1]++;
	      
	      lnt++;
	      v3 = base[v0][v1][v2]->other_p[0];
	      gre03[v0]++;
	      gre03[v1]++;
	      gre03[v2]++;
	      gre03[v3]++;
	      gre13[v0][v1]++;
	      gre13[v0][v2]++;
	      gre13[v0][v3]++;
	      gre13[v1][v0]++;
	      gre13[v1][v2]++;
	      gre13[v1][v3]++;
	      gre13[v2][v0]++;
	      gre13[v2][v1]++;
	      gre13[v2][v3]++;
	      gre13[v3][v0]++;
	      gre13[v3][v1]++;
	      gre13[v3][v2]++;
	      if (base[v0][v1][v2]->other_p[1] != -1) {
		lnt++;
		v3 = base[v0][v1][v2]->other_p[1];
		gre03[v0]++;
		gre03[v1]++;
		gre03[v2]++;
		gre03[v3]++;
		gre13[v0][v1]++;
		gre13[v0][v2]++;
		gre13[v0][v3]++;
		gre13[v1][v0]++;
		gre13[v1][v2]++;
		gre13[v1][v3]++;
		gre13[v2][v0]++;
		gre13[v2][v1]++;
		gre13[v2][v3]++;
		gre13[v3][v0]++;
		gre13[v3][v1]++;
		gre13[v3][v2]++;
	      }
	    }
      if (is_e) {
	lne++;
	gre01[v0]++;
	gre01[v1]++;
      }
    }
    if (is_v)
      lnv++;
  }
  
  if (lnv != nv_p) 
    error_exit(211);
  if (lne != 2*ne_p) 
    error_exit(212);
  if (lnf != 3*2*nf_p) 
    error_exit(213);
  if (lnt != 4*3*2*nt_p) 
    error_exit(214);
  
  for (v0=0; v0<maxnv; v0++) {
    if (gre01[v0] != 2*degree01_p[v0])
      error_exit(221);
    if (gre02[v0] != 3*2*degree02_p[v0])
      error_exit(222);
    if (gre03[v0] != 4*3*2*degree03_p[v0])
      error_exit(223);
    for (v1=0; v1<maxnv; v1++) {
      if (gre12[v0][v1] != 3*2*degree12_p[v0][v1])
	error_exit(224);
      if (gre13[v0][v1] != 4*3*2*degree13_p[v0][v1])
	error_exit(225);
    }
  }
  
  return;
}

void dump_it()
{
  int ilist,nface;
  int v0,v1,v2;

  if (TRUE) {
    for (ilist=0; ilist<nt_a; ilist++)
      fprintf(stderr,"[%d,%d,%d,%d] ",
	      list_a[ilist][0]+1,list_a[ilist][1]+1,list_a[ilist][2]+1,
	      list_a[ilist][3]+1);
    fprintf(stderr,"\n");
    return;
  }

  fprintf(stderr,"\nnv_a = %d, ne_a = %d, nf_a = %d, nt_a = %d\n",
	  nv_a,ne_a,nf_a,nt_a);

  fprintf(stderr,"list of tetrahedra:\n");
  for (ilist=0; ilist<nt_a; ilist++)
    fprintf(stderr,"%2d %2d %2d %2d \n",
	    list_a[ilist][0],list_a[ilist][1],list_a[ilist][2],list_a[ilist][3]);

  fprintf(stderr,"edge matrix:\n");
  fprintf(stderr,"   |");
  for (v1=0; v1<nv_a; v1++) 
    fprintf(stderr,"%2d ",v1);
  fprintf(stderr,"\n");
  fprintf(stderr,"----");
  for (v1=0; v1<nv_a; v1++) 
    fprintf(stderr,"---");
  fprintf(stderr,"\n");
  for (v0=0; v0<nv_a; v0++) {
    fprintf(stderr,"%2d |",v0);
    for (v1=0; v1<nv_a; v1++) 
      if (degree12_a[v0][v1] != 0)
	fprintf(stderr," X ");
      else
	fprintf(stderr,"   ");
    fprintf(stderr,"\n");
  }

  nface = 0;
  fprintf(stderr,"face others:\n");
  for (v0=0; v0<nv_a-2; v0++)
    for (v1=v0+1; v1<nv_a-1; v1++)
      for (v2=v1+1; v2<nv_a; v2++)
	if (base[v0][v1][v2]->other_a[0] != -1) {
	  nface++;
	  fprintf(stderr,"(%d,%d,%d) %d %d\n",
		  v0,v1,v2,
		  base[v0][v1][v2]->other_a[0],base[v0][v1][v2]->other_a[1]);
	}

  fprintf(stderr,"edge degree:");
  for (v1=0; v1<nv_a; v1++) 
    fprintf(stderr,"%2d ",degree01_a[v1]);
  fprintf(stderr,"\n");

  fprintf(stderr,"face degree:");
  for (v1=0; v1<nv_a; v1++) 
    fprintf(stderr,"%2d ",degree02_a[v1]);
  fprintf(stderr,"\n");

  fprintf(stderr,"tetrahedron degree:");
  for (v1=0; v1<nv_a; v1++) 
    fprintf(stderr,"%2d ",degree03_a[v1]);
  fprintf(stderr,"\n");

  if (nface != nf_a) 
    error_exit(33);
}

void add_other_a(int v[3], int vo, int type)
{
  /* add vertex vo to base (v[0],v[1],v[2]) to make tetrahedron */

  int i;

#ifdef COVER

  int nb,b[3],bo;
  
  bo = cm[nt_a][vo];
  if (bo == -1) {
    nb = 0;
    for (i=0; i<3; i++) {
      b[i] = cm[nt_a][v[i]];
      if (b[i] != -1)
	nb++;
    }
    if (nb == 3)
      if (base[v[0]][v[1]][v[2]]->other_a[0] != -1) {
	bo = cm[nt_a][base[v[0]][v[1]][v[2]]->other_a[0]];
	if (bo != -1) {
	  if (base[b[0]][b[1]][b[2]]->other_c[0] == bo)
	    cm[nt_a][vo] = base[b[0]][b[1]][b[2]]->other_c[1];
	  else
	    cm[nt_a][vo] = base[b[0]][b[1]][b[2]]->other_c[0];
	  cmult[nt_a][cm[nt_a][vo]]++;
	}
      }
  }
  
#endif

  if (base[v[0]][v[1]][v[2]]->other_a[0] != -1) {
    base[v[0]][v[1]][v[2]]->other_a[1] = vo;
    base[v[0]][v[1]][v[2]]->type_a[1] = type;
  }
  else {
    base[v[0]][v[1]][v[2]]->other_a[0] = vo;
    base[v[0]][v[1]][v[2]]->type_a[0] = type;
    
    nf_a++;
    
    for (i=0; i<3; i++)
      degree02_a[v[i]]++;

    degree12_a[v[0]][v[1]]++;
    degree12_a[v[1]][v[0]]++;
    degree12_a[v[0]][v[2]]++;
    degree12_a[v[2]][v[0]]++;
    degree12_a[v[1]][v[2]]++;
    degree12_a[v[2]][v[1]]++;
  }
}

void add_other_p(int v[3], int vo)
{
  /* add vertex vo to base (v[0],v[1],v[2]) to make tetrahedron */

  int i;

  if (base[v[0]][v[1]][v[2]]->other_p[0] != -1)
    base[v[0]][v[1]][v[2]]->other_p[1] = vo;
  else {
    base[v[0]][v[1]][v[2]]->other_p[0] = vo;
    
    nf_p++;
    
    for (i=0; i<3; i++)
      degree02_p[v[i]]++;

    degree12_p[v[0]][v[1]]++;
    degree12_p[v[1]][v[0]]++;
    degree12_p[v[0]][v[2]]++;
    degree12_p[v[2]][v[0]]++;
    degree12_p[v[1]][v[2]]++;
    degree12_p[v[2]][v[1]]++;
  }
}

void add_tetrahedron_a(int v[4], int type)
{
  /* add a tetrahedron to 3-manifold */

  int i,j,vloc[7];

  for (i=0;i<4;i++)
    vloc[i] = v[i];
  for (i=0;i<3;i++)
    vloc[i+4] = v[i];

  /* check if previously forced */

  if (type == PICKED)
    if (base[vloc[0]][vloc[1]][vloc[2]]->other_a[0] == vloc[3] ||
	base[vloc[0]][vloc[1]][vloc[2]]->other_a[1] == vloc[3]) {
      for (i=0;i<4;i++)
	if (base[vloc[0+i]][vloc[1+i]][vloc[2+i]]->other_a[0] == vloc[3+i])
	  base[vloc[0+i]][vloc[1+i]][vloc[2+i]]->type_a[0] += PICKED;
	else
	  base[vloc[0+i]][vloc[1+i]][vloc[2+i]]->type_a[1] += PICKED;
      return;
    }
  
  /* update tetrahedron of faces (23) (12) */

  for (i=0;i<4;i++)
    add_other_a(vloc+i,vloc[3+i],type);

  /* update list of tetrahedra (03) */

  for (i=0;i<4;i++) {
    list_a[nt_a][i] = v[i];
    degree03_a[v[i]]++;
  }
  if (type == FORCED)
    list_a_forced[nt_a] = nt_p-1;
  else
    list_a_forced[nt_a] = -1;

  nt_a++;

#ifdef COVER

  for (i=0; i<maxnv; i++)
    cm[nt_a][i] = cm[nt_a-1][i];
  for (i=0; i<nv_c; i++)
    cmult[nt_a][i] = cmult[nt_a-1][i];

#endif

  /* update nv */

  if (v[2] == nv_a)
    nv_a++;
  if (v[3] == nv_a)
    nv_a++;
  if (nv_a > max_inter_nv)
    max_inter_nv = nv_a;

  /* update degrees of vertices (01) and degrees of edges (13) */
  
  for (i=0;i<3;i++)
    for (j=i+1;j<4;j++) {
      if (degree13_a[v[i]][v[j]] == 0) {
	degree01_a[v[i]]++;
	degree01_a[v[j]]++;
	ne_a++;
      }
      degree13_a[v[i]][v[j]]++;
      degree13_a[v[j]][v[i]]++;
    }

  /* update complete */

  for (i=0;i<4;i++)
    if (2*degree02_a[v[i]] == 3*degree03_a[v[i]]) {
      complete_a[v[i]] = TRUE;
      completeness[v[i]] = 1;
    }
  
  //check_it();
}

void add_tetrahedron_p(int v[4], int type)
{
  /* add a tetrahedron to 3-manifold */

  int i,j,vloc[7];

  if (type == PICKED) {
    
    /* update tetrahedron of faces (23) (12) */
    
    for (i=0;i<4;i++)
      vloc[i] = v[i];
    for (i=0;i<3;i++)
      vloc[i+4] = v[i];
    
    for (i=0;i<4;i++)
      add_other_p(vloc+i,vloc[3+i]);
    
    /* update list of tetrahedra (03) */
    
    for (i=0;i<4;i++) {
      list_p[nt_p][i] = v[i];
      degree03_p[v[i]]++;
    }
    nt_p++;
    
    /* update nv */
    
    if (v[2] == nv_p)
      nv_p++;
    if (v[3] == nv_p)
      nv_p++;
    if (nv_p > max_inter_nv)
      max_inter_nv = nv_p;
    
    /* update degrees of vertices (01) and degrees of edges (13) */
    
    for (i=0;i<3;i++)
      for (j=i+1;j<4;j++) {
	if (degree13_p[v[i]][v[j]] == 0) {
	  degree01_p[v[i]]++;
	  degree01_p[v[j]]++;
	  ne_p++;
	}
	degree13_p[v[i]][v[j]]++;
	degree13_p[v[j]][v[i]]++;
      }
    
    /* update complete */
    
    for (i=0;i<4;i++)
      if (2*degree02_p[v[i]] == 3*degree03_p[v[i]])
	complete_p[v[i]] = TRUE;
  }  
  
  add_tetrahedron_a(v,type);

  //check_it();

}

void remove_other_a(int v[3], int vo)
{
  /* remove vertex vo from base (v[0],v[1],v[2]) */
  
  int i;

  if (base[v[0]][v[1]][v[2]]->other_a[1] == vo) {
    base[v[0]][v[1]][v[2]]->other_a[1] = -1;
    base[v[0]][v[1]][v[2]]->type_a[1] = AVAILABLE;
  }
  else {
    base[v[0]][v[1]][v[2]]->other_a[0] = -1;
    base[v[0]][v[1]][v[2]]->type_a[0] = AVAILABLE;
    nf_a--;

    for (i=0; i<3; i++)
      degree02_a[v[i]]--;

    degree12_a[v[0]][v[1]]--;
    degree12_a[v[1]][v[0]]--;
    degree12_a[v[0]][v[2]]--;
    degree12_a[v[2]][v[0]]--;
    degree12_a[v[1]][v[2]]--;
    degree12_a[v[2]][v[1]]--;
  }
}

void remove_other_p(int v[3], int vo)
{
  /* remove vertex vo from base (v[0],v[1],v[2]) */
  
  int i;

  if (base[v[0]][v[1]][v[2]]->other_p[1] == vo) {
    base[v[0]][v[1]][v[2]]->other_p[1] = -1;
  }
  else {
    base[v[0]][v[1]][v[2]]->other_p[0] = -1;
    nf_p--;

    for (i=0; i<3; i++)
      degree02_p[v[i]]--;

    degree12_p[v[0]][v[1]]--;
    degree12_p[v[1]][v[0]]--;
    degree12_p[v[0]][v[2]]--;
    degree12_p[v[2]][v[0]]--;
    degree12_p[v[1]][v[2]]--;
    degree12_p[v[2]][v[1]]--;
  }
}

void remove_tetrahedron_a(int type, int v[4])
{
  int i,j;
  int vloc[7];

  for (i=0;i<4;i++)
    vloc[i] = v[i];
  for (i=0;i<3;i++)
    vloc[i+4] = v[i];
  
  if (type == PICKED)
    if ((base[v[0]][v[1]][v[2]]->other_a[0] == v[3] &&
	 base[v[0]][v[1]][v[2]]->type_a[0] == FORCED_PICKED) ||
	(base[v[0]][v[1]][v[2]]->other_a[1] == v[3] &&
	 base[v[0]][v[1]][v[2]]->type_a[1] == FORCED_PICKED)) {
      
      for (i=0;i<4;i++)
	if (base[vloc[i+0]][vloc[i+1]][vloc[i+2]]->other_a[0] == vloc[i+3])
	  base[vloc[i+0]][vloc[i+1]][vloc[i+2]]->type_a[0] = FORCED;
	else
	  base[vloc[i+0]][vloc[i+1]][vloc[i+2]]->type_a[1] = FORCED;
	
      return;
    }
  
  /* update list */
  
  nt_a--;

  /* update tetrahedron degrees of vertices (03) */

  for (i=0;i<4;i++) {
    degree03_a[v[i]]--;
    complete_a[v[i]] = FALSE;
    completeness[v[i]] = 0;
  }

  /* update nv */

  if (degree03_a[v[3]] == 0)
    nv_a--;
  if (degree03_a[v[2]] == 0)
    nv_a--;

  /* update degrees of vertices (01) and degrees of edges (13) */
  
  for (i=0;i<3;i++)
    for (j=i+1;j<4;j++) {
      degree13_a[v[i]][v[j]]--;
      degree13_a[v[j]][v[i]]--;
      if (degree13_a[v[i]][v[j]] == 0) {
	degree01_a[v[i]]--;
	degree01_a[v[j]]--;
	ne_a--;
      }
    }

  /* update tetrahedron of faces (23) (12) */

  for (i=0;i<4;i++)
    remove_other_a(vloc+i,vloc[i+3]);

  //check_it();
}

void remove_tetrahedron_p(int type)
{
  int i,j;
  int v[4],vloc[7];

  if (type == PICKED) {

    /* update list */
    
    nt_p--;
    
    for (i=0;i<4;i++)
      v[i] = list_p[nt_p][i];
    
    /* update tetrahedron degrees of vertices (03) */
    
    for (i=0;i<4;i++) {
      degree03_p[v[i]]--;
      complete_p[v[i]] = FALSE;
    }
    
    /* update nv */
    
    if (degree03_p[v[3]] == 0)
      nv_p--;
    if (degree03_p[v[2]] == 0)
      nv_p--;
    
    /* update degrees of vertices (01) and degrees of edges (13) */
    
    for (i=0;i<3;i++)
      for (j=i+1;j<4;j++) {
	degree13_p[v[i]][v[j]]--;
	degree13_p[v[j]][v[i]]--;
	if (degree13_p[v[i]][v[j]] == 0) {
	  degree01_p[v[i]]--;
	  degree01_p[v[j]]--;
	  ne_p--;
	}
      }
    
    /* update tetrahedron of faces (23) (12) */
    
    for (i=0;i<4;i++)
      vloc[i] = v[i];
    for (i=0;i<3;i++)
      vloc[i+4] = v[i];

    for (i=0;i<4;i++)
      remove_other_p(vloc+i,vloc[3+i]);

    while (list_a_forced[nt_a-1] == nt_p)
      remove_tetrahedron_a(FORCED,list_a[nt_a-1]);
    remove_tetrahedron_a(PICKED,v);
  }
  else {
    remove_tetrahedron_a(FORCED,list_a[nt_a-1]);
  }

  //check_it();
}

void initialize()
{
  /* one time initialization */

  int v0,v1,v2,iface,lnv;

#ifdef SPLITTEST
  splitcases = 0;
#endif

  debug_count = 0;
  max_inter_nv = 0;
  ZEROBIG(ncalls_min_lex);

  for (lnv=0; lnv<MAXN; lnv++) {
    ZEROBIG(ngenerated[lnv]);
    ZEROBIG(closed_link[lnv]);
  }
  ZEROBIG(ngen_all);
  prev_closed_link = 0;

  iface = 0;
  for (v0=0; v0<maxnv-2; v0++)
    for (v1=v0+1; v1<maxnv-1; v1++)
      for (v2=v1+1; v2<maxnv; v2++) {
	base[v0][v1][v2] = &faces[iface];
	base[v0][v2][v1] = &faces[iface];
	base[v1][v0][v2] = &faces[iface];
	base[v1][v2][v0] = &faces[iface];
	base[v2][v0][v1] = &faces[iface];
	base[v2][v1][v0] = &faces[iface];

	ofacespt[v0][v1][v2] = &ofaces[2*iface];
	ofacespt[v1][v2][v0] = &ofaces[2*iface];
	ofacespt[v2][v0][v1] = &ofaces[2*iface];
	ofacespt[v0][v2][v1] = &ofaces[2*iface+1];
	ofacespt[v1][v0][v2] = &ofaces[2*iface+1];
	ofacespt[v2][v1][v0] = &ofaces[2*iface+1];

	iface++;
      }

  for (v0=0; v0<maxnv; v0++) {
    vertspt[v0] = &verts[v0];
  }

#ifdef COVER

  for (v0=0; v0<maxnv-2; v0++)
    for (v1=v0+1; v1<maxnv-1; v1++) 
      for (v2=v1+1; v2<maxnv; v2++) {
	base[v0][v1][v2]->other_c[0] = -1;
	base[v0][v1][v2]->other_c[1] = -1;
      }

  nv_c = read_lex();

  if (nv_c == 0) {
    fprintf(stderr,"failed to read triangulation to cover.\n");
    exit(0);}

  if (2*nv_c > maxnv) {
    fprintf(stderr,
	    "twice the number of vertices in triangulation to cover (%d) is smaller than nv (%d).\n",
	    nv_c,maxnv);
    exit(0);}
  
  if ((maxnv % nv_c) != 0) {
    fprintf(stderr,
	    "nv (%d) is not multiple of number of vertices in triangulation to cover (%d).\n",
	    maxnv,nv_c);
    maxnv = nv_c * (maxnv/nv_c);
    fprintf(stderr,
	    "reducing nv to %d.\n",
	    maxnv);
  }

  multi_c = maxnv/nv_c;

  for (v0=0; v0<degree0_c+2; v0++) {
    cm[0][v0] = v0;
    cmult[0][v0] = 1;
  }
  for (v0=degree0_c+2; v0<maxnv; v0++) {
    cm[0][v0] = -1;
    cmult[0][v0] = 0;
  }

#endif
}

void initialize_first_edge(int degree0)

{

  int i,v0,v1,v2,v[4];

  /* clear data*/

  for (v0=0; v0<maxnv-2; v0++)
    for (v1=v0+1; v1<maxnv-1; v1++) 
      for (v2=v1+1; v2<maxnv; v2++) {
	base[v0][v1][v2]->other_a[0] = -1;
	base[v0][v1][v2]->other_a[1] = -1;
	base[v0][v1][v2]->type_a[0] = AVAILABLE;
	base[v0][v1][v2]->type_a[1] = AVAILABLE;
	base[v0][v1][v2]->other_p[0] = -1;
	base[v0][v1][v2]->other_p[1] = -1;
      }

  for (v0=0; v0<maxnv; v0++)
    for (v1=0; v1<maxnv; v1++) {
      degree12_a[v0][v1] = 0;
      degree13_a[v0][v1] = 0;
      degree12_p[v0][v1] = 0;
      degree13_p[v0][v1] = 0;
    }

  for (v0=0; v0<maxnv; v0++) {
    degree01_a[v0] = 0;
    degree02_a[v0] = 0;
    degree03_a[v0] = 0;
    degree01_p[v0] = 0;
    degree02_p[v0] = 0;
    degree03_p[v0] = 0;
    complete_a[v0] = FALSE;
    complete_p[v0] = FALSE;
  }

  for (v0=0; v0<maxnv; v0++)
    v_to_label[v0] = -1;

  nv_a = 3;
  ne_a = 0;
  nf_a = 0;
  nt_a = 0;
  nv_p = 3;
  ne_p = 0;
  nf_p = 0;
  nt_p = 0;

  /* add tetrahedra around first edge */

  for (i=0;i<4;i++)
    v[i] = i;
  add_tetrahedron_p(v,PICKED);
  for (i=4; i<degree0+2; i++) {
    v[2] = i-2;
    v[3] = i;
    add_tetrahedron_p(v,PICKED);
  }
  v[2] = degree0;
  add_tetrahedron_p(v,PICKED);

} 

int mark_face_np(int *lnf, int *lnv, int *lnb, int v0, int v1, int v2, int v3)
{
  /* recursively mark the faces of the link of v0 adjacent to (v1,v2,v3) 
   across (v1,v3) and (v3,v2).
   update lnf, lnv, and lnb.
   if face is found with opposite orientation return TRUE.
 */

  int v4;

  if (!ISMARKED(vertspt[v3])) {
    MARK(vertspt[v3]);
    (*lnv)++;
  }

  if (base[v0][v1][v3]->other_a[0]==v2)
    v4 = base[v0][v1][v3]->other_a[1];
  else
    v4 = base[v0][v1][v3]->other_a[0];
  if (v4 == -1)
    (*lnb)++;
  else if (ISMARKED(ofacespt[v3][v1][v4])) 
    return TRUE;
  else if (!ISMARKED(ofacespt[v1][v3][v4])) {
    MARK(ofacespt[v1][v3][v4]);
    (*lnf)++;
    if (mark_face_np(lnf,lnv,lnb,v0,v1,v3,v4))
      return TRUE;
  }
  
  if (base[v0][v3][v2]->other_a[0]==v1)
    v4 = base[v0][v3][v2]->other_a[1];
  else
    v4 = base[v0][v3][v2]->other_a[0];
  if (v4 == -1)
    (*lnb)++;
  else if (ISMARKED(ofacespt[v2][v3][v4])) 
    return TRUE;
  else if (!ISMARKED(ofacespt[v3][v2][v4])) {
    MARK(ofacespt[v3][v2][v4]);
    (*lnf)++;
    if (mark_face_np(lnf,lnv,lnb,v0,v3,v2,v4))
      return TRUE;
  }

  return FALSE;
}

int nonplanar(int v0, int v1, int v2, int v3)
{
  /* return TRUE if we can figure out that the strongly connected component of
     the link of v0 containing (v1,v2,v3) is not planar.  
     here a sphere is also nonplanar.
  */
  /* for the strongly connected component containing (v1,v2,v3) of link of v0 
     after adding (v0,v1,v2,v3) count
     nf, nv, nb (= number of edges on boundary edges).
     also check that it is orientable.
  */

  int lnf,lnv,lnb;
  int v12,v31,v23;
  int nf_lnb;  /* number of faces required to fill in boundary */

  RESETMARKS;

  /* mark the faces around (v1,v2,v3) first since (v1,v2,v3) has not been added
     yet */

  lnf = 1;
  lnv = 3;
  lnb = 0;

  MARK(vertspt[v1]);
  MARK(vertspt[v2]);
  MARK(vertspt[v3]);

  v12 = base[v0][v1][v2]->other_a[0];
  if (v12 == -1)
    lnb++;
  else {
    lnf++;
    MARK(ofacespt[v1][v2][v12]);
  }

  v23 = base[v0][v2][v3]->other_a[0];
  if (v23 == -1)
    lnb++;
  else {
    lnf++;
    MARK(ofacespt[v2][v3][v23]);
  }

  v31 = base[v0][v3][v1]->other_a[0];
  if (v31 == -1)
    lnb++;
  else {
    lnf++;
    MARK(ofacespt[v3][v1][v31]);
  }

  if (v12 != -1)
    if (mark_face_np(&lnf,&lnv,&lnb,v0,v1,v2,v12))
      return TRUE;
  if (v23 != -1)
    if (mark_face_np(&lnf,&lnv,&lnb,v0,v2,v3,v23))
      return TRUE;
  if (v31 != -1)
    if (mark_face_np(&lnf,&lnv,&lnb,v0,v3,v1,v31))
      return TRUE;

  if (lnb == 0)
    return TRUE;
  nf_lnb = (lnb+2)/3;
  return (lnf + nf_lnb > 2*(lnv-2));
}

#if defined FLAG

int four_connected(int v0, int v1, int v2, int v3)
{
  /* 
     return TRUE if the link of v0 is 4-connected.
     the link is complete except of face (v1,v2,v3) which is about to be added.
     the link is planar.
     so if any triangle (other than (v1,v2,v3)) is not a face return FALSE.
  */

  int v4,v5,v6,nnei;

  for (v4=0; v4<nv_a-1; v4++)
    if (v4 != v0)
      if (degree12_a[v0][v4] > 0)
	for (v5=v4+1; v5<nv_a; v5++)
	  if (v5 != v0)
	    if (base[v0][v4][v5]->other_a[0] != -1) {
	      
	      /* (v4,v5) is an edge in the link of v0 */
	      
	      nnei = 0;
	      for (v6=0; v6<nv_a; v6++)
		if (v6 != v0 && v6 != v4 && v6 != v5)
		  if (base[v0][v4][v6]->other_a[0] != -1 &&
		      base[v0][v5][v6]->other_a[0] != -1)
		    if (++nnei > 2)
		      return FALSE;
	    }
  return TRUE;
}

#endif

int admissable_vertex(int v0, int v1, int v2, int v3)
{
  /* check if adding (v0,v1,v2,v3) would cause the link of v0 to be 
     non-planar */

  int nv_v0, ne_v0, nf_v0;

  nv_v0 = degree01_a[v0];
  if (degree12_a[v0][v1] == 0)
    nv_v0++;
  if (degree12_a[v0][v2] == 0)
    nv_v0++;
  if (degree12_a[v0][v3] == 0)
    nv_v0++;

#if defined DODEC

  if (v0 == 0 && nv_v0 > 12)
    return FALSE;

#endif

#if defined TCP

  if (nv_v0 > MAX_VERTEX_DEGREE)
    return FALSE;

#endif

  nf_v0 = degree03_a[v0]+1;

  if (nf_v0 < 2*(nv_v0 - 2)) {
    
    /* check the 2D-connected component (after tetrahedron is added) */
    
    if (nonplanar(v0,v1,v2,v3))
      return FALSE;
    else
      return TRUE;
  }
  
  else {
    
    /* v0 should be complete */

#if defined NEIGHBORLY

    if (degree01_a[v0] != maxnv-1)
      return FALSE;

#endif

    ne_v0 = degree02_a[v0];
    if (base[v0][v1][v2]->other_a[0] == -1)
      ne_v0++;
    if (base[v0][v1][v3]->other_a[0] == -1)
      ne_v0++;
    if (base[v0][v2][v3]->other_a[0] == -1)
      ne_v0++;
    if (2*ne_v0 == 3*nf_v0) {
      
#if defined TCP

      if (degree01_a[v0] < MIN_VERTEX_DEGREE)
	return FALSE;

#endif

#if defined FLAG

      if (degree01_a[v0] >= 9)
	if (!four_connected(v0,v1,v2,v3))
	  return FALSE;

#endif

      /* link of v0 has no boundary */
      /* link of v0 has Euler characteristic 2 */
      /* link of v0 is connected from tests when previous tetrahrdra 
	 were added */

      return TRUE;
    }

    return FALSE;
  }
}

int admissable_edge(int v0, int v1, int v2, int v3)
{
  /* check if adding (v0,v1,v2,v3) would cause the link of (v0,v1) to be 
     a cycle and at least one other edge */
  
  int va,vb,vc,vd;

#if defined COVER

  if (cm[nt_a][v0] != -1 && cm[nt_a][v1] != -1) {
    if (cm[nt_a][v0] == cm[nt_a][v1])
      fprintf(stderr,"ooops\n");
    if (degree13_a[v0][v1] >= degree13_c[cm[nt_a][v0]][cm[nt_a][v1]])
      return FALSE;
  }

#endif

  /* non pseudo-manifold conditions */

  if (degree12_a[v0][v1] < 2)
    return TRUE;

  if (degree12_a[v0][v1] == degree13_a[v0][v1]+1)
    return TRUE;
  
  if (base[v0][v1][v2]->other_a[0] == -1)
    return TRUE;
  else
    va = base[v0][v1][v2]->other_a[0];
  
  if (base[v0][v1][v3]->other_a[0] == -1)
    return TRUE;
  else
    vb = base[v0][v1][v3]->other_a[0];
  
  vc = v2;
  while (va != vb && va != -1) {
    if (base[v0][v1][va]->other_a[0] == vc)
      vd = base[v0][v1][va]->other_a[1];
    else
      vd = base[v0][v1][va]->other_a[0];
    vc = va;
    va = vd;
  }
  
  if (va == vb)
    return FALSE;
  else
    return TRUE;
}

int smaller_lex()
{
  /* check if the relabeling starting at nt_match is lexigraphically 
     smaller than the current labeling */
  
  int v;
  int label[4];        /* vertices of tetrahedron. also labels in relabeling */
  int w0,w1,w2,w3;      /* vertices of tetrahedron relabeled 
			   (label[0],label[1],label[2],label[3]) */
  int i;                /* vertex index for tetrahedron */
  int branch;           /* search might branch */
  int save_next_label;
  int save_nt_match;

  branch = FALSE;

  for (i=0; i<4; i++)
    label[i] = list_p[nt_match-1][i];
  
  while (nt_match<nt_p) {
    
    w3 = -1;
    while (w3 == -1) 

      /* label[0], label[1], label[2], and label[3] are the vertices of the 
	 nt_match-1 or better */
      
      if (label[0] == list_p[nt_match][0] && 
	  label[1] == list_p[nt_match][1] && 
	  label[2] == list_p[nt_match][2]) {
	
	/* current labeling has another tetrahedron with 
	   (label[0],label[1],label[2]). */
	
	/* find the next relabeled tetrahedron */
	
	w0 = label_to_v[label[0]];
	w1 = label_to_v[label[1]];
	
	if (label[2] == next_label) {
	  
	  /* try all possibilities of assigning next_label to unlabeled 
	     vertices */
	  
	  for (w2=0; w2<nv_a; w2++)
	    if (v_to_label[w2] == -1) 
	      if (base[w0][w1][w2]->other_a[0] != -1) {
		v_to_label[w2] = next_label;
		label_to_v[next_label] = w2;
		save_next_label = next_label;
		next_label++;
		save_nt_match = nt_match;
		if (smaller_lex())
		  return TRUE;
		for ( ; next_label>save_next_label; next_label--)
		  v_to_label[label_to_v[next_label-1]] = -1;
		nt_match = save_nt_match;
	      }
	  return FALSE;
	}
	
	w2 = label_to_v[label[2]];
	w3 = -1;
	
	if (base[w0][w1][w2]->other_a[0] == -1)
	  return FALSE;
	
	/* (w0,w1,w2) exists */
	
	if (base[w0][w1][w2]->other_a[1] != -1) {
	  
	  /* (w0,w1,w2) is complete */
	  
	  if (v_to_label[base[w0][w1][w2]->other_a[0]] > label[3] &&
	      v_to_label[base[w0][w1][w2]->other_a[1]] > label[3]) {
	    
	    /* both labels of other vertices could be next so pick smaller */
	    
	    if (v_to_label[base[w0][w1][w2]->other_a[0]] > 
		v_to_label[base[w0][w1][w2]->other_a[1]])
	      w3 = base[w0][w1][w2]->other_a[1];
	    else
	      w3 = base[w0][w1][w2]->other_a[0];
	  }
	  else if (v_to_label[base[w0][w1][w2]->other_a[0]] > label[3])
	    
	    /* one label of other vertices could be next so use it */
	    
	    w3 = base[w0][w1][w2]->other_a[0];
	  else if (v_to_label[base[w0][w1][w2]->other_a[1]] > label[3])
	    w3 = base[w0][w1][w2]->other_a[1];
	  else if (v_to_label[base[w0][w1][w2]->other_a[0]] == -1) {
	    if (v_to_label[base[w0][w1][w2]->other_a[1]] == -1)
	      
	      /* both final vertices of the tetrahedra of (w0,w1,w2) are 
		 unlabeled.  the search must branch below */
	      
	      branch = TRUE;
	    
	    /* first final vertex of the tetrahedra of (w0,w1,w2) is
	       unlabeled.  use it. */
	    
	    w3 = base[w0][w1][w2]->other_a[0];
	    v_to_label[w3] = next_label;
	    label_to_v[next_label] = w3;
	    next_label++;
	  }
	  else if (v_to_label[base[w0][w1][w2]->other_a[1]] == -1) {
	    
	    /* second final vertex of the tetrahedra of (w0,w1,w2) is only
	       unlabeled.  use it. */
	    
	    w3 = base[w0][w1][w2]->other_a[1];
	    v_to_label[w3] = next_label;
	    label_to_v[next_label] = w3;
	    next_label++;
	  }
	}
	else {
	  
	  /* (w0,w1,w2) exists but is not complete */
	  
	  if (v_to_label[base[w0][w1][w2]->other_a[0]] > label[3])
	    w3 = base[w0][w1][w2]->other_a[0];
	  else if (v_to_label[base[w0][w1][w2]->other_a[0]] == -1) {
	    w3 = base[w0][w1][w2]->other_a[0];
	    v_to_label[w3] = next_label;
	    label_to_v[next_label] = w3;
	    next_label++;
	  }
	}
	
	/* compare labels */
	
	if (w3 == -1)
	  return FALSE;
	if (v_to_label[w3] > list_p[nt_match][3])
	  return FALSE;
	if (v_to_label[w3] < list_p[nt_match][3])
	  return TRUE;

	/* current labeling and relabeling both have another tetrahedron 
	   with (label[0],label[1],label[2]).  they are the same. */
	
	if (branch) {
	  save_nt_match = nt_match;
	  nt_match++;
	  save_next_label = next_label-1;
	  if (smaller_lex())
	    return TRUE;
	  for ( ; next_label>save_next_label; next_label--)
	    v_to_label[label_to_v[next_label-1]] = -1;
	  nt_match = save_nt_match;
	  branch = FALSE;
	  w3 = base[w0][w1][w2]->other_a[1];
	  v_to_label[w3] = next_label;
	  label_to_v[next_label] = w3;
	  next_label++;
	}
	  
	label[3] = v_to_label[w3];
      }
    
      else {
	
	/* current labeling does not have another tetrahedron with 
	   (label[0],label[1],label[2]). */
	
	/* find the next relabeled tetrahedron */
	
	w0 = label_to_v[label[0]];
	w1 = label_to_v[label[1]];
	w3 = -1;
	
	if (label[2] == next_label) {
	  
	  /* try all possibilities of assigning next_label to unlabeled 
	     vertices */
	  
	  for (w2=0; w2<nv_a; w2++)
	    if (v_to_label[w2] == -1) 
	      if (base[w0][w1][w2]->other_a[0] != -1)
		return TRUE;
	}
	else {
	  
	  w2 = label_to_v[label[2]];
	  
	  if (base[w0][w1][w2]->other_a[0] != -1) {
	    
	    /* (w0,w1,w2) exists */
	    
	    if (v_to_label[base[w0][w1][w2]->other_a[0]] > label[3] ||
		v_to_label[base[w0][w1][w2]->other_a[0]] == -1) 
	      return TRUE;
	    
	    if (base[w0][w1][w2]->other_a[1] != -1)
	      if (v_to_label[base[w0][w1][w2]->other_a[1]] > label[3] ||
		  v_to_label[base[w0][w1][w2]->other_a[1]] == -1) 
		return TRUE;
	  }
	}
	
	/* neither current labeling nor relabeling have another tetrahedron 
	   with (label[0],label[1],label[2]). increase this 3-tuple */
	
	label[2]++;
	if (label[2] == next_label+1 ||
	    label[2] == nv_a-1) {
	  label[1]++;
	  if (label[1] == next_label ||
	      label[1] == nv_a-2) {
	    label[0]++;
	    label[1] = label[0] + 1;
	  }
	  label[2] = label[1] + 1;
	}
	label[3] = label[2];
      }
    
    nt_match++;
  }
  
  /* they are the same */
  
  return FALSE;
}

int min_lex()
{
  /* check that the current labeling of the list of tetrahedra is 
     lexigraphical minimum */

  int v0,v1,v,i;
  int ineighbor,ioffset,save;
  int neighbor[MAXN];
  int save_next_label;
  int imatch;

  ADDBIG(ncalls_min_lex,1);

  save_next_label = degree12_a[0][1]+1+1;

  for (v0=0; v0<nv_a; v0++) {
    if (completeness[v0] == 1)
      nmatch[v0] = 0;

#if defined DODEC
    if (completeness[0] != 2 || (complete_a[v0] && degree01_a[v0] == 12))
#endif
    if (completeness[0] != 2 || completeness[v0] != 2) {
      for (v1=0; v1<nv_a; v1++) 
	if (v1 != v0)
	  if (degree13_a[v0][v1] == degree13_a[0][1] &&
	      degree13_a[v0][v1] == degree12_a[v0][v1]) {
	    
	    /* make a list of the neighbors of (v0,v1) */
	    
	    for (v=0; (v==v0||v==v1?TRUE:base[v0][v1][v]->other_a[0] == -1); v++)
	      {}
	    neighbor[0] = v;
	    neighbor[1] = base[v0][v1][v]->other_a[0];
	    for (ineighbor=2; ineighbor<degree12_a[0][1]; ineighbor++)
	      if (base[v0][v1][neighbor[ineighbor-1]]->other_a[0] == 
		  neighbor[ineighbor-2])
		neighbor[ineighbor] = 
		  base[v0][v1][neighbor[ineighbor-1]]->other_a[1];
	      else
		neighbor[ineighbor] = 
		  base[v0][v1][neighbor[ineighbor-1]]->other_a[0];
	    
	    /* set the initial values of v_to_label for (v0,v1) */
	    
	    v_to_label[v0] = 0;
	    label_to_v[0] = v0;
	    v_to_label[v1] = 1;
	    label_to_v[1] = v1;
	    v_to_label[v] = 2;
	    label_to_v[2] = v;
	    for (ioffset=1; ioffset<=degree12_a[0][1]/2; ioffset++) {
	      v_to_label[neighbor[ioffset]] = 2*ioffset+1;
	      label_to_v[2*ioffset+1] = neighbor[ioffset];
	    }
	    for (ioffset=1; ioffset<(degree12_a[0][1]+1)/2; ioffset++) {
	      v_to_label[neighbor[-ioffset+degree12_a[0][1]]] = 2*ioffset+1+1;
	      label_to_v[2*ioffset+1+1] = neighbor[-ioffset+degree12_a[0][1]];
	    }
	    for (ineighbor=0; ineighbor<degree12_a[0][1]; ineighbor++) {
	      
	      /* check starting at each neighbor of (v0,v1) going clockwise */
	      
	      nt_match = degree12_a[0][1];
	      next_label = save_next_label;
	      if (smaller_lex()) {
		for ( ; next_label>0; next_label--)
		  v_to_label[label_to_v[next_label-1]] = -1;
		return FALSE;
	      }

	      for ( ; next_label>save_next_label; next_label--)
		v_to_label[label_to_v[next_label-1]] = -1;
	      if (completeness[v0] == 1)
		if (nt_match >= degree03_a[0]) {
		  for (v=0; v<=degree01_a[0]; v++)
		    match[v0][nmatch[v0]][v] = label_to_v[v];
		  nmatch[v0]++;
		}
	      
	      /* rotate the labeling for next neighbor */
	      
	      save = v_to_label[neighbor[degree12_a[0][1]-1]];
	      for (ioffset=degree12_a[0][1]-1; ioffset>0; ioffset--) {
		v_to_label[neighbor[ioffset]] = 
		  v_to_label[neighbor[ioffset-1]];
		label_to_v[v_to_label[neighbor[ioffset]]] = neighbor[ioffset];
	      }
	      v_to_label[neighbor[0]] = save;
	      label_to_v[v_to_label[neighbor[0]]] = neighbor[0];
	    }
	    
	    /* flip the labeling for counterclockwise check */
	    
	    for (ioffset=1; ioffset<(degree12_a[0][1]+1)/2; ioffset++) {
	      save = v_to_label[neighbor[ioffset]];
	      v_to_label[neighbor[ioffset]] = 
		v_to_label[neighbor[-ioffset+degree12_a[0][1]]];
	      label_to_v[v_to_label[neighbor[ioffset]]] = neighbor[ioffset];
	      v_to_label[neighbor[-ioffset+degree12_a[0][1]]] = save;
	      label_to_v[v_to_label[neighbor[-ioffset+degree12_a[0][1]]]] = 
		neighbor[-ioffset+degree12_a[0][1]];
	    }
	    
	    for (ineighbor=0; ineighbor<degree12_a[0][1]; ineighbor++) {
	      
	      /* check starting at each neighbor of (v0,v1) going 
		 counterclockwise */
	      
	      nt_match = degree12_a[0][1];
	      next_label = save_next_label;
	      if (smaller_lex()) {
		for ( ; next_label>0; next_label--)
		  v_to_label[label_to_v[next_label-1]] = -1;
		return FALSE;
	      }
	      
	      for ( ; next_label>save_next_label; next_label--)
		v_to_label[label_to_v[next_label-1]] = -1;
	      if (completeness[v0] == 1)
		if (nt_match >= degree03_a[0]) {
		  for (v=0; v<=degree01_a[0]; v++)
		    match[v0][nmatch[v0]][v] = label_to_v[v];
		  nmatch[v0]++;
		}
	      
	      /* rotate the labeling for next neighbor */
	      
	      save = v_to_label[neighbor[degree12_a[0][1]-1]];
	      for (ioffset=degree12_a[0][1]-1; ioffset>0; ioffset--) {
		v_to_label[neighbor[ioffset]] = 
		  v_to_label[neighbor[ioffset-1]];
		label_to_v[v_to_label[neighbor[ioffset]]] = neighbor[ioffset];
	      }
	      v_to_label[neighbor[0]] = save;
	      label_to_v[v_to_label[neighbor[0]]] = neighbor[0];
	    }
	    
	    /* undo labeling before next (v0,v1) or returning */
	    
	    for ( ; next_label>0; next_label--)
	      v_to_label[label_to_v[next_label-1]] = -1;
	  }
    }
    
    else if (nmatch[v0] != 0) {
      
      /* completeness[0] == 2 && completeness[v0] == 2 
	 and lk(0) is equivalent to lk(v0) */
      
      for (imatch=0; imatch<nmatch[v0]; imatch++) {
	for (v=0; v<=degree01_a[0]; v++) {
	  label_to_v[v] = match[v0][imatch][v];
	  v_to_label[label_to_v[v]] = v;
	}
	nt_match = degree03_a[0];
	next_label = degree01_a[0] + 1;
	if (smaller_lex()) {
	  for ( ; next_label>0; next_label--)
	    v_to_label[label_to_v[next_label-1]] = -1;
	  return FALSE;
	}
	for ( ; next_label>degree01_a[0]+1; next_label--)
	  v_to_label[label_to_v[next_label-1]] = -1;
      }
      for ( ; next_label>0; next_label--)
	v_to_label[label_to_v[next_label-1]] = -1;
    }

    if (complete_p[0] && completeness[v0] == 1)
      completeness[v0] = 2;
    
  }

  return TRUE;
}

int contractible(int v0, int v1)
{
  /* check to see if the edge (v0,v1) is contractible.  vertices v0 and v1 are
     complete. */

  int v2,v3;
  int neighbors;  /* number of common neighbors of v0 and v1 */
  int eneighbors;  /* number of common neighboring edges of v0 and v1 */

  neighbors = 0;
  eneighbors = 0;
  for (v2=0; v2<nv_a; v2++)
    if (v2!=v0 && v2!=v1)
      if (degree12_a[v0][v2] != 0 && degree12_a[v1][v2] != 0) {
	neighbors++;
	if (neighbors == degree12_a[v0][v1]+1)
	  return FALSE;
	for (v3=v2+1; v3<nv_a; v3++)
	  if (v3!=v0 && v3!=v1)
	    if (base[v0][v2][v3]->other_a[0] != -1 && 
		base[v1][v2][v3]->other_a[0] != -1) {
	      eneighbors++;
	      if (eneighbors == degree12_a[v0][v1]+1)
		return FALSE;
	    }
      }
  return TRUE;
}

void found_one()
{
  /* found a triangulation */
  
  char outfilename[100];
  int ilist;
  
  ADDBIG(ngenerated[nv_a],1);
  ADDBIG(ngen_all,1);
  
  if (FALSE) {
    if (outfile_alpha[ne_a] == NULL) {
      sprintf(outfilename,"v%d_e%d_r%s_m%d.alpha\0",
	      nv_a,ne_a,res_text,mod);
      if ((outfile_alpha[ne_a] = fopen(outfilename,"w")) == NULL)
	{
	  fprintf(stderr,
		  "can't open %s for writing\n",outfilename);
	  exit(1);
	}
    }
    /*    
	  write_alpha(outfile_alpha[ne_a]);
    */
  }
  
  if (outfile_lex[nv_a] == NULL) {
    sprintf(outfilename,"3-manifolds_v%d_r%s_m%d.lex\0",
	    nv_a,res_text,mod);
    if ((outfile_lex[nv_a] = fopen(outfilename,"w")) == NULL)
      {
	fprintf(stderr,
		"can't open %s for writing\n",outfilename);
	exit(1);
      }
  }
  
  write_lex(outfile_lex[nv_a]);
  fflush(outfile_lex[nv_a]);
}

int complete_edge(int v[4])
{
  /* find the other two vertices of the tetrahedron which would close the 
     link of (v[0],v[1]) */

  int va,vb,vc;

  va = base[v[0]][v[1]][v[2]]->other_a[0];
  vb = v[2];
  while (va != -1) {
    if (base[v[0]][v[1]][va]->other_a[0] == vb)
      vc = base[v[0]][v[1]][va]->other_a[1];
    else
      vc = base[v[0]][v[1]][va]->other_a[0];
    vb = va;
    va = vc;
  }
  v[3] = vb;
  
  va = base[v[0]][v[1]][v[2]]->other_a[1];
  vb = v[2];
  while (va != -1) {
    if (base[v[0]][v[1]][va]->other_a[0] == vb)
      vc = base[v[0]][v[1]][va]->other_a[1];
    else
      vc = base[v[0]][v[1]][va]->other_a[0];
    vb = va;
    va = vc;
  }
  v[2] = vb;

  /* check that tetrahedron can be added */

  if (base[v[0]][v[2]][v[3]]->other_a[1] != -1 ||
      base[v[1]][v[2]][v[3]]->other_a[1] != -1)
    return FALSE;
  
  return TRUE;
}

int force_tetrahedron(int v[4])
{
  /* check if the just added tetrahedron forces additional tetrahedra to be
     added.
     if forced tetrahedron lead to a conflict, return FALSE.  
  */

  int vloc[4];
  int i,j,k;
  int vk;

#if defined MAX5

  /* known: degree13_a <= degree12_a <= 5 */

  for (i=0; i<3; i++)
    for (j=i+1; j<4; j++)
      if (degree13_a[v[i]][v[j]] == 4 && degree12_a[v[i]][v[j]] == 5) {
	
	/* force the tetrahedron which completes (v[i],v[j]) */
	
	for (k=0; k==i||k==j; k++) {}
	vloc[0] = v[i];
	vloc[1] = v[j];
	vloc[2] = v[k];
	if (!complete_edge(vloc))
	  return FALSE;
	if (!do_tetrahedron(vloc,FORCED))
	  return FALSE;
      }
  
#endif  
  
#if defined NO66

  /* known: degree13_a <= degree12_a <= 6 and
     no face of new tetrahedron has more than 1 edge with degree12_a == 6 */

  for (i=0; i<3; i++)
    for (j=i+1; j<4; j++)
      if (degree12_a[v[i]][v[j]] == 6) {
	if (degree13_a[v[i]][v[j]] == 5) {
	  
	  /* force the tetrahedron which completes (v[i],v[j]) */
	  
	  for (k=0; k==i||k==j; k++) {}
	  vloc[0] = v[i];
	  vloc[1] = v[j];
	  vloc[2] = v[k];
	  if (!complete_edge(vloc))
	    return FALSE;
	  if (!do_tetrahedron(vloc,FORCED))
	    return FALSE;
	}
	
	/* force any tetrahedron which completes a degree 5 edge on a face 
	   with (v[i],v[j]) */
	
	for (vk=0; vk<nv_a; vk++)
	  if (vk != v[i] && vk != v[j])
	    if (base[vk][v[i]][v[j]]->type_a[0] != AVAILABLE) {
	      if (degree12_a[vk][v[i]] == 5 && degree13_a[vk][v[i]] == 4) {
		vloc[0] = vk;
		vloc[1] = v[i];
		vloc[2] = v[j];
		if (!complete_edge(vloc))
		  return FALSE;
		if (!do_tetrahedron(vloc,FORCED))
		  return FALSE;
	      }
	      if (degree12_a[vk][v[j]] == 5 && degree13_a[vk][v[j]] == 4) {
		vloc[0] = vk;
		vloc[1] = v[j];
		vloc[2] = v[i];
		if (!complete_edge(vloc))
		  return FALSE;
		if (!do_tetrahedron(vloc,FORCED))
		  return FALSE;
	      }
	    }
      }
      else if (degree12_a[v[i]][v[j]] == 5)
	if (degree13_a[v[i]][v[j]] == 4)
	  
	  /* if (v[i],v[j]) is on a face with an edge of degree 6 force the
	     completion of (v[i],v[j]) */
	  
	  for (vk=0; vk<nv_a; vk++)
	    if (vk != v[i] && vk != v[j])
	      if (base[vk][v[i]][v[j]]->type_a[0] != AVAILABLE)
		if (degree12_a[vk][v[i]] == 6 || degree12_a[vk][v[j]] == 6) {
		  for (k=0; k==i||k==j; k++) {}
		  vloc[0] = v[i];
		  vloc[1] = v[j];
		  vloc[2] = v[k];
		  if (!complete_edge(vloc))
		    return FALSE;
		  if (!do_tetrahedron(vloc,FORCED))
		    return FALSE;
		  break;
		}
  
#endif  
  
#if defined COVER

  /* final edge degrees will be the same as in the base triangulation */
 
  int b[4];

  for (i=0; i<4; i++)
    b[i] = cm[nt_a][v[i]];
   
  for (i=0; i<3; i++)
    if (b[i] != -1)
      for (j=i+1; j<4; j++) 
	if (b[j] != -1)
	  if (degree13_a[v[i]][v[j]] == degree13_c[b[i]][b[j]] - 1 &&
	      degree12_a[v[i]][v[j]] == degree13_c[b[i]][b[j]]) {

	    /* force the tetrahedron which completes (v[i],v[j]) */
	    
	    for (k=0; k==i||k==j; k++) {}
	    vloc[0] = v[i];
	    vloc[1] = v[j];
	    vloc[2] = v[k];
	    if (!complete_edge(vloc))
	      return FALSE;
	    if (!do_tetrahedron(vloc,FORCED))
	      return FALSE;
	  }
  
#endif  

#if defined TEST

  /* SIMPLE TEST: force in tetrahedron [1,3,5,6] */

  if (maxnv > 6)
    if (base[1][3][5]->other_a[0] != 6 &&
	base[1][3][5]->other_a[1] != 6) {
      vloc[0] = 1;
      vloc[1] = 3;
      vloc[2] = 5;
      vloc[3] = 6;
      return do_tetrahedron(vloc,FORCED);
    }
  
#endif

  /* no conflict */
  
  return TRUE;
}

int do_tetrahedron(int v[4], int type)
{
  int done;
  int vi,vj,vk,orient;
  int i,j,check_vi;
  int vloc[7];

  for (i=0;i<4;i++)
    vloc[i] = v[i];
  for (i=0;i<3;i++)
    vloc[i+4] = v[i];

#if defined NO66
  int n66,e66[2][2],i66;
#endif

  /* add a new tetrahedron and check pruning conditions. */

  /* check if new picked tetrahedron was previously forced */

  if (type == PICKED &&
      (base[v[0]][v[1]][v[2]]->other_a[0] == v[3] ||
       base[v[0]][v[1]][v[2]]->other_a[1] == v[3])) {
    add_tetrahedron_p(v,type);
    if (!min_lex()) {
      remove_tetrahedron_p(type);
      return FALSE;
    }
  }

  else {

    /* pruning before adding tetrahedron */
    
#ifdef COVER

    int nb,b[4];

    /* check if tetrahedron maps (or could map) onto the base triangulation */

    nb = 0;
    for (i=0; i<4; i++) {
      b[nb] = cm[nt_a][v[i]];
      if (b[nb] != -1)
	nb++;
    }
    if (nb == 4) {
      if (b[0] == b[1] || b[0] == b[2] || b[0] == b[3] ||
	  b[1] == b[2] || b[1] == b[3] || b[2] == b[3])
	return FALSE;
      if (base[b[0]][b[1]][b[2]]->other_c[0] != b[3] &&
	  base[b[0]][b[1]][b[2]]->other_c[1] != b[3])
	return FALSE;
      if (base[v[0]][v[1]][v[2]]->other_a[0] != -1)
	if (cm[nt_a][base[v[0]][v[1]][v[2]]->other_a[0]] == b[3])
	  return FALSE;
      if (base[v[0]][v[1]][v[3]]->other_a[0] != -1)
	if (cm[nt_a][base[v[0]][v[1]][v[3]]->other_a[0]] == b[2])
	  return FALSE;
      if (base[v[0]][v[2]][v[3]]->other_a[0] != -1)
	if (cm[nt_a][base[v[0]][v[2]][v[3]]->other_a[0]] == b[1])
	  return FALSE;
      if (base[v[1]][v[2]][v[3]]->other_a[0] != -1)
	if (cm[nt_a][base[v[1]][v[2]][v[3]]->other_a[0]] == b[0])
	  return FALSE;
    }
    else if (nb == 3) {
      if (b[0] == b[1] || b[0] == b[2] || b[1] == b[2])
	return FALSE;
      if (base[b[0]][b[1]][b[2]]->other_c[0] == -1)
	return FALSE;
    }
    else if (nb == 2) {
      if (b[0] == b[1])
	return FALSE;
      if (degree13_c[b[0]][b[1]] == 0)
	return FALSE;
    }

    //    else
    //fprintf(stderr,"nb = %d\n",nb);

    /* temporary code ????? */

    int ilist;

    for (ilist=0; ilist<nt_a; ilist++) {
      nb = 0;
      for (i=0; i<4; i++) {
	b[i] = cm[nt_a][list_a[ilist][i]];
	if (b[i] != -1)
	  nb++;
      }
      if (nb == 4) {
	if (b[0] == b[1] || b[0] == b[2] || b[0] == b[3] ||
	    b[1] == b[2] || b[1] == b[3] || b[2] == b[3]) {
/* 	  fprintf(stderr,"1nt_a,ilist,list_a,b = %d,%d,(%d,%d,%d,%d),(%d,%d,%d,%d)\n",nt_a,ilist,list_a[ilist][0],list_a[ilist][1],list_a[ilist][2],list_a[ilist][3],b[0],b[1],b[2],b[3]); */
	  return FALSE;
	}
	if (base[b[0]][b[1]][b[2]]->other_c[0] != b[3] &&
	    base[b[0]][b[1]][b[2]]->other_c[1] != b[3]) {
/* 	  fprintf(stderr,"2nt_a,ilist,list_a,b = %d,%d,(%d,%d,%d,%d),(%d,%d,%d,%d)\n",nt_a,ilist,list_a[ilist][0],list_a[ilist][1],list_a[ilist][2],list_a[ilist][3],b[0],b[1],b[2],b[3]); */
	  return FALSE;
	}
	if (base[b[0]][b[1]][b[2]]->other_c[0] == b[3] &&
	    base[b[0]][b[1]][b[2]]->other_c[1] == b[3]) {
/* 	  fprintf(stderr,"3nt_a,ilist,list_a,b = %d,%d,(%d,%d,%d,%d),(%d,%d,%d,%d)\n",nt_a,ilist,list_a[ilist][0],list_a[ilist][1],list_a[ilist][2],list_a[ilist][3],b[0],b[1],b[2],b[3]); */
	  return FALSE;
	}
      }
    }

#endif

    /* check if addition of tetrahedron is blocked by forced tetrahedra */
    
    if (type == PICKED)
      for (i=0; i<4; i++)
	if (base[vloc[i+0]][vloc[i+1]][vloc[i+2]]->other_a[1] != -1)
	  if (base[vloc[i+0]][vloc[i+1]][vloc[i+2]]->other_a[0] != vloc[i+3] &&
	      base[vloc[i+0]][vloc[i+1]][vloc[i+2]]->other_a[1] != vloc[i+3])
	    return FALSE;
    
    /* check if link of vertex or link of edge is already complete */

    for (i=0; i<4; i++)
      if (complete_a[v[i]])
	return FALSE;
    for (i=0; i<4-1; i++)
      for (j=i+1; j<4; j++) 
	if (degree12_a[v[i]][v[j]] != 0 &&
	    degree12_a[v[i]][v[j]] == degree13_a[v[i]][v[j]])
	  return FALSE;

    /* check that adding tetrahedron would not cause the link of (v[i],v[j]) 
       to be a cycle and at least one other edge */
    
    if (!admissable_edge(v[0],v[1],v[2],v[3]))
      return FALSE;
    if (!admissable_edge(v[0],v[2],v[1],v[3]))
      return FALSE;
    if (!admissable_edge(v[0],v[3],v[1],v[2]))
      return FALSE;
    if (!admissable_edge(v[1],v[2],v[0],v[3]))
      return FALSE;
    if (!admissable_edge(v[1],v[3],v[0],v[2]))
      return FALSE;
    if (!admissable_edge(v[2],v[3],v[0],v[1]))
      return FALSE;
    
    if (!admissable_vertex(v[0],v[1],v[2],v[3]))
      return FALSE;
    if (!admissable_vertex(v[1],v[0],v[2],v[3]))
      return FALSE;
    if (!admissable_vertex(v[2],v[0],v[1],v[3]))
      return FALSE;
    if (!admissable_vertex(v[3],v[0],v[1],v[2]))
      return FALSE;
    
    /* add a new tetrahedron */
    
    add_tetrahedron_p(v,type);
    
    /* check that (0,1) still has minimal degree of complete edges */
    
    for (i=0; i<4-1; i++)
      for (j=i+1; j<4; j++) {
	if (degree12_a[v[i]][v[j]] == degree13_a[v[i]][v[j]] &&
	    degree12_a[v[i]][v[j]] < degree12_a[0][1]) {
	  remove_tetrahedron_p(type);
	  return FALSE;
	}
      }
    
#if defined MAX5
    
    for (i=0; i<4-1; i++)
      for (j=i+1; j<4; j++) {
	if (degree12_a[v[i]][v[j]] > MAX_EDGE_DEGREE) {
	  remove_tetrahedron_p(type);
	  return FALSE;
	}
      }
    
#endif
    
#if defined NEIGHBORLY
    
    if (only_irreducible)
      for (i=0; i<4-1; i++)
	for (j=i+1; j<4; j++) {
	  if (degree12_a[v[i]][v[j]] > maxnv - 3) {
	    remove_tetrahedron_p(type);
	    return FALSE;
	  }
	}
    
#endif
    
#if defined NO66
    
    n66 = 0;
    for (i=0; i<4-1; i++)
      for (j=i+1; j<4; j++) {
	if (degree12_a[v[i]][v[j]] > MAX_EDGE_DEGREE) {
	  remove_tetrahedron_p(type);
	  return FALSE;
	}
	if (degree12_a[v[i]][v[j]] == 6) {
	  n66++;
	  if (n66 > 2) {
	    remove_tetrahedron_p(type);
	    return FALSE;
	  }
	  if (n66 == 2) {
	    if (i == e66[0][0] || i == e66[0][1] || j == e66[0][1]) {
	      remove_tetrahedron_p(type);
	      return FALSE;
	    }
	  }
	  e66[n66-1][0] = i;
	  e66[n66-1][1] = j;
	}
      }
    
    for (i66=0; i66<n66; i66++) {
      vi = v[e66[i66][0]];
      vj = v[e66[i66][1]];
      for (vk=0; vk<nv_a; vk++)
	if (vk != vi && vk != vj)
	  if (base[vi][vj][vk]->other_a[0] != -1)
	    if (degree12_a[vi][vk] == 6 ||
		degree12_a[vj][vk] == 6) {
	      remove_tetrahedron_p(type);
	      return FALSE;
	    }
    }
    
#endif
    
#if !defined NEIGHBORLY
    
    /* check for contractible edge */
    
    if (only_irreducible && nv_a > 5) 
      for (i=0; i<4; i++) {
	if (complete_a[v[i]])
	  for (vi=0; vi<nv_a; vi++) {
	    if (degree12_a[v[i]][vi] != 0) {
	      check_vi = TRUE;
	      for (j=0; check_vi && j<i; j++)
		check_vi = (vi != v[j]);
	      if (check_vi)
		if (complete_a[vi])
		  if (contractible(v[i],vi)) {
		    remove_tetrahedron_p(type);
		    return FALSE;
		  }
	    }
	  }
      }
    
#endif
    
#ifdef COVER
    
    /* check if any vertex of the just added tetrahedron requires too many 
       vertices to map onto a single vertex of the triangulataion to be covered */
    
    for (i=0; i<4; i++)
      if (cmult[nt_a][cm[nt_a][v[i]]] > multi_c){
	remove_tetrahedron_p(type);
	return FALSE;
      }
    
#endif
    
    /* check if the just added tetrahedron forces additional tetrahedra to be
       added */
    
    if (!force_tetrahedron(v)) {
      remove_tetrahedron_p(type);
      return FALSE;
    }

    /* pruning after adding tetrahedron */
    
    if (type == PICKED)
      if (!min_lex()) {
	remove_tetrahedron_p(type);
	return FALSE;
      }
    
  }

  /* check triangulation is done */
    
  if (type == PICKED) {
    done = TRUE;
    for (vi=v[0]; done && vi<nv_p; vi++)
      done = (complete_p[vi]);

    if (vi > prev_closed_link)
      ADDBIG(closed_link[vi],1);
    prev_closed_link = vi;

    if (done) {
      
      //if (nv == maxnv) {
      
      found_one();
      //}
      
      /*
	dump_it();
      */
      
      remove_tetrahedron_p(type);
      return TRUE;
    }
    
    /* do it again */
    
    next_tetrahedron();
    
    remove_tetrahedron_p(type);
  }

  return TRUE;
}

void next_tetrahedron()
{
  /* a tetrahedron has just been added.  
     find what the next tetrahedron might be. */

  int v[4];
  int v2lim,v3lim;

  if (nt_p == splitlevel) {
#ifdef SPLITTEST
    splitcases++;
    return;
#endif
    if (splitcount-- != 0) return;
    splitcount = mod - 1;
  }

  v[0] = list_p[nt_p-1][0];
  v[1] = list_p[nt_p-1][1];
  v[2] = list_p[nt_p-1][2];
  v[3] = list_p[nt_p-1][3]+1;

  /* check if v[0] is complete */

  if (complete_p[v[0]]) {
    while (complete_p[v[0]])
      v[0]++;
    v[1] = v[0]+1;
    v[2] = v[1]+1;
    v[3] = v[2]+1;
  }

  /* check if (v[0],v[1]) is complete */
  
  if (degree12_p[v[0]][v[1]] == degree13_p[v[0]][v[1]] &&
      degree12_p[v[0]][v[1]] != 0) {
    while (degree12_p[v[0]][v[1]] == degree13_p[v[0]][v[1]] &&
      degree12_p[v[0]][v[1]] != 0)
      v[1]++;
    v[2] = v[1]+1;
    v[3] = v[2]+1;
  }

#if defined NEIGHBORLY
  
  /* while (v[0],v[1]) is complete (NEIGHBORLY) */
  
  while (degree12_p[v[0]][v[1]] != 0 &&
	 degree12_p[v[0]][v[1]] == degree13_p[v[0]][v[1]]) {
    v[1]++;
    v[2] = v[1]+1;
    v[3] = v[2]+1;
  }
  
  
  /* now (v[0],v[1]) is not complete (NEIGHBORLY) */
  
  /* if (v[0],v[1]) is unused (NEIGHBORLY) */
  
  if (degree12_p[v[0]][v[1]] == 0) {
    
    /* first use of (v[0],v[1]) */
    
    v2lim = MIN(nv_p+1,maxnv-1);
    for (v[2]=v[1]+1; v[2]<v2lim; v[2]++)
      if ((degree12_p[v[0]][v[2]] == 0 || 
	   degree12_p[v[0]][v[2]] != degree13_p[v[0]][v[2]]) &&
	  (degree12_p[v[1]][v[2]] == 0 || 
	   degree12_p[v[1]][v[2]] != degree13_p[v[1]][v[2]]) &&
	  (v[2] == nv_p || degree03_p[v[2]] != 2*(degree01_p[v[2]]-2))) {
	v3lim = MIN(MAX(nv_p,v[2]+1)+1,maxnv);
	for (v[3]=v[2]+1; v[3]<v3lim; v[3]++)
	  if ((degree12_p[v[0]][v[3]] == 0 || 
	       degree12_p[v[0]][v[3]] != degree13_p[v[0]][v[3]]) &&
	      (degree12_p[v[1]][v[3]] == 0 || 
	       degree12_p[v[1]][v[3]] != degree13_p[v[1]][v[3]]) &&
	      (degree12_p[v[2]][v[3]] == 0 || 
	       degree12_p[v[2]][v[3]] != degree13_p[v[2]][v[3]]) &&
	      (base[v[1]][v[2]][v[3]]->other_p[1] == -1) &&
	      (base[v[0]][v[2]][v[3]]->other_p[1] == -1) &&
	      (v[3] >= nv_p || !complete_p[v[3]]))
	    do_tetrahedron(v,PICKED);
      }
  }
  
  else {
    
    /* (v[0],v[1]) is on the boundary (NEIGBORLY) */
    
    /* while (v[0],v[1],v[2]) is not on boundry */
    
    while (v[2] < maxnv-1 &&
	   (base[v[0]][v[1]][v[2]]->other_p[0] == -1 ||
	    base[v[0]][v[1]][v[2]]->other_p[1] != -1)) {
      if (base[v[0]][v[1]][v[2]]->other_p[0] == -1 &&
	  !complete_p[v[2]]) {
	
	/* first use of (v[0],v[1],v[2]) */
	
	v3lim = MIN(MAX(nv_p,v[2]+1)+1,maxnv);
	for (v[3]=v[2]+1; v[3]<v3lim; v[3]++)
	  if ((base[v[0]][v[1]][v[3]]->other_p[1] == -1) &&
	      (base[v[0]][v[2]][v[3]]->other_p[1] == -1) &&
	      (base[v[1]][v[2]][v[3]]->other_p[1] == -1) &&
	      (v[3] == nv_p || !complete_p[v[3]]))
	    do_tetrahedron(v,PICKED);
      }
      v[2]++;
      v[3] = v[2]+1;
    }
    
    /* (v[0],v[1],v[2]) is on boundry.  must be used */
    
    v3lim = MIN(MAX(nv_p,v[2]+1)+1,maxnv);
    for (; v[3]<v3lim; v[3]++)
      if ((base[v[0]][v[1]][v[3]]->other_p[1] == -1) &&
	  (base[v[0]][v[2]][v[3]]->other_p[1] == -1) &&
	  (base[v[1]][v[2]][v[3]]->other_p[1] == -1) &&
	  (v[3] == nv_p || degree03_p[v[3]] != 2*(degree01_p[v[3]]-2)))
	do_tetrahedron(v,PICKED);
  }
  
#endif
  
#if ! defined NEIGHBORLY
  
  /* while (v[0],v[1]) is not on boundry */
  
  while (degree12_p[v[0]][v[1]] == 0 ||
	 degree12_p[v[0]][v[1]] == degree13_p[v[0]][v[1]]) {
    if (degree12_p[v[0]][v[1]] == 0 && !complete_p[v[1]]) {
      
      /* first use of (v[0],v[1]) */
      
      v2lim = MIN(nv_p+1,maxnv-1);
      for (v[2]=v[1]+1; v[2]<v2lim; v[2]++)
	if ((degree12_p[v[0]][v[2]] == 0 || 
	     degree12_p[v[0]][v[2]] != degree13_p[v[0]][v[2]]) &&
	    (degree12_p[v[1]][v[2]] == 0 || 
	     degree12_p[v[1]][v[2]] != degree13_p[v[1]][v[2]]) &&
	    (v[2] == nv_p || !complete_p[v[2]])) {
	  v3lim = MIN(MAX(nv_p,v[2]+1)+1,maxnv);
	  for (v[3]=v[2]+1; v[3]<v3lim; v[3]++)
	    if ((degree12_p[v[0]][v[3]] == 0 || 
		 degree12_p[v[0]][v[3]] != degree13_p[v[0]][v[3]]) &&
		(degree12_p[v[1]][v[3]] == 0 || 
		 degree12_p[v[1]][v[3]] != degree13_p[v[1]][v[3]]) &&
		(degree12_p[v[2]][v[3]] == 0 || 
		 degree12_p[v[2]][v[3]] != degree13_p[v[2]][v[3]]) &&
		(base[v[1]][v[2]][v[3]]->other_p[1] == -1) &&
		(base[v[0]][v[2]][v[3]]->other_p[1] == -1) &&
		(v[3] >= nv_p || !complete_p[v[3]]))
	      do_tetrahedron(v,PICKED);
	}
    }
    v[1]++;
    v[2] = v[1]+1;
    v[3] = v[2]+1;
  }

  /* while (v[0],v[1],v[2]) is not on boundry */

  while (v[2] < maxnv-1 &&
	 (base[v[0]][v[1]][v[2]]->other_p[0] == -1 ||
	  base[v[0]][v[1]][v[2]]->other_p[1] != -1)) {
    if (base[v[0]][v[1]][v[2]]->other_p[0] == -1 && !complete_p[v[2]]) {
      
      /* first use of (v[0],v[1],v[2]) */

      v3lim = MIN(MAX(nv_p,v[2]+1)+1,maxnv);
      for (v[3]=v[2]+1; v[3]<v3lim; v[3]++)
	if ((base[v[0]][v[1]][v[3]]->other_p[1] == -1) &&
	    (base[v[0]][v[2]][v[3]]->other_p[1] == -1) &&
	    (base[v[1]][v[2]][v[3]]->other_p[1] == -1) &&
	    (v[3] == nv_p || !complete_p[v[3]]))
	  do_tetrahedron(v,PICKED);
    }
    v[2]++;
    v[3] = v[2]+1;
  }

  /* (v[0],v[1],v[2]) is on boundry.  must be used */
  
  v3lim = MIN(MAX(nv_p,v[2]+1)+1,maxnv);
  for (; v[3]<v3lim; v[3]++)
    if ((base[v[0]][v[1]][v[3]]->other_p[1] == -1) &&
	(base[v[0]][v[2]][v[3]]->other_p[1] == -1) &&
	(base[v[1]][v[2]][v[3]]->other_p[1] == -1) &&
	(v[3] == nv_p || !complete_p[v[3]]))
      do_tetrahedron(v,PICKED);

#endif

}
 
static void
initialize_splitting(int minlevel, int hint, int maxlevel)
  
/* Set splitlevel and splitcount.  */
{
  splitlevel = hint;
  if (splitlevel > maxlevel) splitlevel = maxlevel;
  
  if (splitlevel < minlevel && splitlevel > 0)
    {
      if (minlevel <= maxlevel) splitlevel = minlevel;
      else                      splitlevel = 0;
    }
  if (mod == 1) splitlevel = 0;
  splitcount = res;
}

static int 
getargvalue(char *arg)
  
/* Find integer value of arguement. 
   arg is a pointer to a command-line argument.
   The value of the arguement is the function return value.
   An absent value is equivalent to 0.
*/
{
  int j,ans,neg;
  
  ans = 0;
  if (arg[0] == '-') {
    neg = -1;
    j = 1;
  }
  else {
    neg = +1;
    j = 0;
  }
  
  for (; arg[j] >= '0' && arg[j] <= '9'; ++j)
    ans = ans * 10 + (arg[j] - '0');
  
  if (arg[j] != '\0')
    return -999999;
  else
    return neg*ans;
}

void summarize() 
{
  /* summarize results */

  int lnv;

  if (verbose) {
    if (!ISZEROBIG(ngen_all)) {
      PRINTBIG(stderr,ngen_all);
      fprintf(stderr," generated, total\n");
    }
    for (lnv=0; lnv<MAXN; lnv++)
      if (!ISZEROBIG(ngenerated[lnv])) {
	PRINTBIG(stderr,ngenerated[lnv]);
	fprintf(stderr," generated, %d vertices\n",lnv);
      }
    for (lnv=0; lnv<MAXN; lnv++)
      if (!ISZEROBIG(closed_link[lnv])) {
	PRINTBIG(stderr,closed_link[lnv]);
	fprintf(stderr," times %d vertex links closed\n",lnv);
      }
  }
  
  fprintf(stderr,"maximum value nv obtained = %d\n",max_inter_nv);

  PRINTBIG(stderr,ncalls_min_lex);
  fprintf(stderr," calls to min_lex\n");
}

main(argc,argv)
int argc;
char *argv[];

{
  int iarg;
  int badargs,gdefined;
  char *arg;
  int orient;
  int hint;
  int degree0,min_degree0,max_degree0;

#if CPUTIME
  struct tms timestruct0,timestruct1;
  
  times(&timestruct0);
#endif
 
  gdefined = FALSE;
  only_irreducible = FALSE;
  orientability = -1;
  maxnv = 0;
  res = 0;
  mod = 1;
  badargs = FALSE;

  for (iarg = 0; iarg < argc ; ++iarg) fprintf(stderr,"%s ",argv[iarg]);
  fprintf(stderr,"\n");

  iarg = 1;
  while (iarg < argc) {
    arg = argv[iarg];
    if (strcmp(arg,"-v") == 0) {
      verbose = TRUE;
      iarg++;
    }
    else if (strcmp(arg,"-h") == 0) {
      fprintf(stderr,"%s\n",HELPTEXT);
      exit(1);
    }
    else if (strcmp(arg,"-r") == 0) {
      iarg++;
      if (iarg >= argc) {
	fprintf(stderr,"res value required after -r switch\n");
	badargs = TRUE;
      }
      else {
	res = getargvalue(argv[iarg]);
	iarg++;
	if (res < 0) {
	  fprintf(stderr,"res (%d) must be non-negative.\n",res);
	  badargs = TRUE;
	  res = 0;
	}
      }
    }
    else if (strcmp(arg,"-m") == 0) {
      iarg++;
      if (iarg >= argc) {
	fprintf(stderr,"mod value required after -m switch\n");
	badargs = TRUE;
      }
      else {
	mod = getargvalue(argv[iarg]);
	iarg++;
	if (mod < 1) {
	  fprintf(stderr,"mod (%d) must be posotive.\n",mod);
	  badargs = TRUE;
	  mod = 1;
	}
      }
    }
    else if (strcmp(arg,"-o") == 0) {
      if (orientability == 0 || orientability == 1) {
	fprintf(stderr,"o switch is redefining orientability\n");
	badargs = TRUE;
      }
      iarg++;
      if (iarg >= argc) {
	fprintf(stderr,"value required after -o switch\n");
	badargs = TRUE;
      }
      else {
	orientability = getargvalue(argv[iarg]);
	iarg++;
	if (orientability != 0 && orientability != 1) {
	  fprintf(stderr,"orientability (%d) must be 0 or 1.\n",
		  orientability);
	  badargs = TRUE;
	  orientability = -1;
	}
      }
    }
    else if (strcmp(arg,"-i") == 0) {
      only_irreducible = TRUE;
      iarg++;
    }
    else {
      if (maxnv > 0) {
	fprintf(stderr,"redefining nv\n");
	badargs = TRUE;
      }
      else
	maxnv = getargvalue(arg);
      iarg++;
      if (maxnv == -999999) {
	fprintf(stderr,"unknown switch %s\n",arg);
	badargs = TRUE;
      }
      if (maxnv < 4) {
	fprintf(stderr,"nv (%d) too small.\n",maxnv);
	badargs = TRUE;
      }
    }
  }

  if (maxnv == 0) {
    fprintf(stderr,"nv not specified\n");
    badargs = TRUE;
  }

  if (maxnv > MAXN) {
    fprintf(stderr,"nv (%d) too large.  increase MAXN.\n",maxnv);
    badargs = TRUE;
  }

  if (res >= mod) {
    fprintf(stderr,"res (%d) >= mod (%d).\n",res,mod);
    badargs = TRUE;
    res = 0;
    mod = 1;
  }

  maxne = maxnv*(maxnv-1)/2;

  if (maxne > MAXE) {
    fprintf(stderr,
	    "maximum number of edges (%d) too large. increase MAXE.\n",
	    maxne);
    badargs = TRUE;
  }

  maxnf = maxnv*(maxnv-1)*(maxnv-2)/(3*2);

  if (maxnf > MAXF) {
    fprintf(stderr,
	    "maximum number of edges (%d) too large. increase MAXF.\n",
	    maxnf);
    badargs = TRUE;
  }

  maxnt = maxne - maxnv;

  if (maxnt > MAXT) {
    fprintf(stderr,
	    "maximum number of tetrahedra (%d) too large. increase MAXT.\n",
	    maxnt);
    badargs = TRUE;
  }

  if (badargs) {
    fprintf(stderr,"Usage: %s\n",USAGE);
    exit(1);
  }
  
#ifdef SPLITTEST
  if (mod == 1) mod = 2;
#endif

  if (verbose) {
    fprintf(stderr,"Version %s\n",VERSION);
    fprintf(stderr,"Number of vertices: %d\n", maxnv);
    fprintf(stderr,"Maximum number of tetrahedra: %d\n", maxnt);
    if (orientability == 0)
      fprintf(stderr,"Only non-orientable are generated\n");
    else if (orientability == 1)
      fprintf(stderr,"Only orientable are generated\n");
    else
      fprintf(stderr,"Either non-orientable or orientable are generated\n");
    if (only_irreducible)
      fprintf(stderr,"Only irreducible triangulations are generated.\n");
    fprintf(stderr,"res/mod = %d/%d\n",res,mod);
  }

  hint = MIN(maxnt/2, 54);

#if defined NO66

  hint = MIN(maxnt/2, 30);

#endif

#if defined FLAG

  hint = MIN(hint,22);

#endif

  initialize_splitting(4,hint,maxnt-1);

  if (verbose && mod != 1)
    fprintf(stderr,"splitlevel is at %d tetrahedra.\n",splitlevel);

  sprintf(res_text,"%d",mod-1);
  sprintf(res_text,"%c%d.%dd",'%',strlen(res_text),strlen(res_text));
  sprintf(res_text,res_text,res);

#if defined TCP

  if (verbose)
    fprintf(stderr,"%d <= vertex degree <= %d, %d <= edge degree <= %d\n",
	    MIN_VERTEX_DEGREE,MAX_VERTEX_DEGREE,
	    MIN_EDGE_DEGREE,MAX_EDGE_DEGREE);

#endif

#if defined DODEC

  if (verbose)
    fprintf(stderr,"At least on vertex has vertex degree 12.\n");

#endif

  initialize();

  min_degree0 = 3;
  max_degree0 = MIN(maxnv-2,5);
  if (only_irreducible)
    max_degree0 = MIN(maxnv-3,max_degree0);

#if defined TCP

  if (min_degree0 < MIN_EDGE_DEGREE)
    min_degree0 = MIN_EDGE_DEGREE;
  if (max_degree0 > MAX_EDGE_DEGREE)
    max_degree0 = MAX_EDGE_DEGREE;

#endif

#if defined COVER

  min_degree0 = max_degree0 = degree0_c;

#endif

#if defined FLAG

  min_degree0 = 4;

#endif

  for (degree0=min_degree0; degree0 <= max_degree0; degree0++) {
    
    initialize_first_edge(degree0);
    
    next_tetrahedron();
  }

#if CPUTIME
  times(&timestruct1);
#endif
  
#ifdef SPLITTEST
  fprintf(stderr,"%d splitting cases at level = %d",splitcases,splitlevel);
#if CPUTIME
  fprintf(stderr,"; cpu=%.2f sec\n",
	  (double)(timestruct1.tms_utime+timestruct1.tms_stime
		   -timestruct0.tms_utime-timestruct0.tms_stime) / 
	  (double)CLK_TCK);
#else
  fprintf(stderr,"\n");
#endif
  return 0;
#endif
  
  summarize();
  
#if CPUTIME
  fprintf(stderr,"cpu=%.2f sec\n",
	  (double)(timestruct1.tms_utime+timestruct1.tms_stime
		   -timestruct0.tms_utime+timestruct0.tms_stime) / 
	  (double)CLK_TCK);
#else
  fprintf(stderr,"\n");
#endif

  exit(0);
}
